import os
import threading
import time as ttime
import warnings
from typing import List

import numpy as np
from bec_lib import MessageEndpoints, bec_logger, messages
from ophyd import Component as Cpt
from ophyd import Device, DeviceStatus
from ophyd import DynamicDeviceComponent as Dcpt
from ophyd import OphydObject, PositionerBase, Signal
from ophyd.sim import EnumSignal, SynSignal, _ReadbackSignal, _SetpointSignal
from ophyd.utils import LimitError, ReadOnlyError

logger = bec_logger.logger


class DeviceStop(Exception):
    pass


class SynSignalRO(Signal):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._metadata.update(
            write_access=False,
        )

    def wait_for_connection(self, timeout=0):
        super().wait_for_connection(timeout)
        self._metadata.update(connected=True)

    def get(self, **kwargs):
        self._readback = np.random.rand()
        return self._readback


class _ReadbackSignal(Signal):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._metadata.update(
            connected=True,
            write_access=False,
        )

    def get(self):
        self._readback = self.parent.sim_state["readback"]
        self.parent.sim_state["readback_ts"] = ttime.time()
        return self._readback

    def describe(self):
        res = super().describe()
        # There should be only one key here, but for the sake of
        # generality....
        for k in res:
            res[k]["precision"] = self.parent.precision
        return res

    @property
    def timestamp(self):
        """Timestamp of the readback value"""
        return self.parent.sim_state["readback_ts"]

    def put(self, value, *, timestamp=None, force=False):
        raise ReadOnlyError("The signal {} is readonly.".format(self.name))

    def set(self, value, *, timestamp=None, force=False):
        raise ReadOnlyError("The signal {} is readonly.".format(self.name))


class _ReadbackSignalCompute(Signal):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._metadata.update(
            connected=True,
            write_access=False,
        )

    def get(self):
        readback = self.parent._compute()
        self._readback = self.parent.sim_state["readback"] = readback
        return readback

    def describe(self):
        res = super().describe()
        # There should be only one key here, but for the sake of
        # generality....
        for k in res:
            res[k]["precision"] = self.parent.precision
        return res

    @property
    def timestamp(self):
        """Timestamp of the readback value"""
        return self.parent.sim_state["readback_ts"]

    def put(self, value, *, timestamp=None, force=False):
        raise ReadOnlyError("The signal {} is readonly.".format(self.name))

    def set(self, value, *, timestamp=None, force=False):
        raise ReadOnlyError("The signal {} is readonly.".format(self.name))


class _ReadbackSignalRand(Signal):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._metadata.update(
            connected=True,
            write_access=False,
        )

    def get(self):
        self._readback = np.random.rand()
        return self._readback

    def describe(self):
        res = super().describe()
        # There should be only one key here, but for the sake of
        # generality....
        for k in res:
            res[k]["precision"] = self.parent.precision
        return res

    @property
    def timestamp(self):
        """Timestamp of the readback value"""
        return self.parent.sim_state["readback_ts"]

    def put(self, value, *, timestamp=None, force=False):
        raise ReadOnlyError("The signal {} is readonly.".format(self.name))

    def set(self, value, *, timestamp=None, force=False):
        raise ReadOnlyError("The signal {} is readonly.".format(self.name))


class _SetpointSignal(Signal):
    def put(self, value, *, timestamp=None, force=False):
        self._readback = float(value)
        self.parent.set(float(value))

    def get(self):
        self._readback = self.parent.sim_state["setpoint"]
        return self.parent.sim_state["setpoint"]

    def describe(self):
        res = super().describe()
        # There should be only one key here, but for the sake of generality....
        for k in res:
            res[k]["precision"] = self.parent.precision
        return res

    @property
    def timestamp(self):
        """Timestamp of the readback value"""
        return self.parent.sim_state["setpoint_ts"]


class _IsMovingSignal(Signal):
    def put(self, value, *, timestamp=None, force=False):
        self._readback = float(value)
        self.parent.set(float(value))

    def get(self):
        self._readback = self.parent.sim_state["is_moving"]
        return self.parent.sim_state["is_moving"]

    def describe(self):
        res = super().describe()
        # There should be only one key here, but for the sake of generality....
        for k in res:
            res[k]["precision"] = self.parent.precision
        return res

    @property
    def timestamp(self):
        """Timestamp of the readback value"""
        return self.parent.sim_state["is_moving_ts"]


class SynAxisMonitor(Device):
    """
    A synthetic settable Device mimic any 1D Axis (position, temperature).

    Parameters
    ----------
    name : string, keyword only
    readback_func : callable, optional
        When the Device is set to ``x``, its readback will be updated to
        ``f(x)``. This can be used to introduce random noise or a systematic
        offset.
        Expected signature: ``f(x) -> value``.
    value : object, optional
        The initial value. Default is 0.
    delay : number, optional
        Simulates how long it takes the device to "move". Default is 0 seconds.
    precision : integer, optional
        Digits of precision. Default is 3.
    parent : Device, optional
        Used internally if this Signal is made part of a larger Device.
    kind : a member the Kind IntEnum (or equivalent integer), optional
        Default is Kind.normal. See Kind for options.
    """

    readback = Cpt(_ReadbackSignalRand, value=0, kind="hinted")

    SUB_READBACK = "readback"
    _default_sub = SUB_READBACK

    def __init__(
        self,
        *,
        name,
        readback_func=None,
        value=0,
        delay=0,
        precision=3,
        parent=None,
        labels=None,
        kind=None,
        **kwargs,
    ):
        if readback_func is None:

            def readback_func(x):
                return x

        sentinel = object()
        loop = kwargs.pop("loop", sentinel)
        if loop is not sentinel:
            warnings.warn(
                f"{self.__class__} no longer takes a loop as input.  "
                "Your input will be ignored and may raise in the future",
                stacklevel=2,
            )
        self.sim_state = {}
        self._readback_func = readback_func
        self.delay = delay
        self.precision = precision
        self.tolerance = kwargs.pop("tolerance", 0.5)

        # initialize values
        self.sim_state["readback"] = readback_func(value)
        self.sim_state["readback_ts"] = ttime.time()

        super().__init__(name=name, parent=parent, labels=labels, kind=kind, **kwargs)
        self.readback.name = self.name


class _SLSDetectorConfigSignal(Signal):
    def put(self, value, *, timestamp=None, force=False):
        self._readback = value
        self.parent.sim_state[self.name] = value

    def get(self):
        self._readback = self.parent.sim_state[self.name]
        return self.parent.sim_state[self.name]


class SynSLSDetector(Device):
    USER_ACCESS = []
    exp_time = Cpt(_SLSDetectorConfigSignal, name="exp_time", value=1, kind="config")
    file_path = Cpt(_SLSDetectorConfigSignal, name="file_path", value="", kind="config")
    file_pattern = Cpt(_SLSDetectorConfigSignal, name="file_pattern", value="", kind="config")
    frames = Cpt(_SLSDetectorConfigSignal, name="frames", value=1, kind="config")
    burst = Cpt(_SLSDetectorConfigSignal, name="burst", value=1, kind="config")
    save_file = Cpt(_SLSDetectorConfigSignal, name="save_file", value=False, kind="config")

    def __init__(self, *, name, kind=None, parent=None, device_manager=None, **kwargs):
        self.device_manager = device_manager
        super().__init__(name=name, parent=parent, kind=kind, **kwargs)
        self.sim_state = {
            f"{self.name}_file_path": "~/Data10/data/",
            f"{self.name}_file_pattern": f"{self.name}_{{:05d}}.h5",
            f"{self.name}_frames": 1,
            f"{self.name}_burst": 1,
            f"{self.name}_save_file": False,
            f"{self.name}_exp_time": 0,
        }
        self._stopped = False
        self.file_name = ""
        self.metadata = {}

    def trigger(self):
        status = DeviceStatus(self)

        self.subscribe(status._finished, event_type=self.SUB_ACQ_DONE, run=False)

        def acquire():
            try:
                for _ in range(self.burst.get()):
                    ttime.sleep(self.exp_time.get())
                    if self._stopped:
                        raise DeviceStop
            except DeviceStop:
                pass
            finally:
                self._stopped = False
                self._done_acquiring()

        threading.Thread(target=acquire, daemon=True).start()
        return status

    def stage(self) -> List[object]:
        msg = self.device_manager.producer.get(MessageEndpoints.scan_status())
        scan_msg = messages.ScanStatusMessage.loads(msg)
        self.metadata = {
            "scanID": scan_msg.content["scanID"],
            "RID": scan_msg.content["info"]["RID"],
            "queueID": scan_msg.content["info"]["queueID"],
        }
        scan_number = scan_msg.content["info"]["scan_number"]
        self.frames.set(scan_msg.content["info"]["num_points"])
        self.file_name = os.path.join(
            self.file_path.get(), self.file_pattern.get().format(scan_number)
        )
        return super().stage()

    def unstage(self) -> List[object]:
        signals = {"config": self.sim_state, "data": self.file_name}
        msg = messages.DeviceMessage(signals=signals, metadata=self.metadata)
        self.device_manager.producer.set_and_publish(
            MessageEndpoints.device_read(self.name), msg.dumps()
        )
        return super().unstage()

    def stop(self, *, success=False):
        super().stop(success=success)
        self._stopped = True


class DummyController:
    USER_ACCESS = [
        "some_var",
        "controller_show_all",
        "_func_with_args",
        "_func_with_args_and_kwargs",
        "_func_with_kwargs",
        "_func_without_args_kwargs",
    ]
    some_var = 10
    another_var = 20

    def on(self):
        self._connected = True

    def off(self):
        self._connected = False

    def _func_with_args(self, *args):
        return args

    def _func_with_args_and_kwargs(self, *args, **kwargs):
        return args, kwargs

    def _func_with_kwargs(self, **kwargs):
        return kwargs

    def _func_without_args_kwargs(self):
        return None

    def controller_show_all(self):
        """dummy controller show all

        Raises:
            in: _description_
            LimitError: _description_

        Returns:
            _type_: _description_
        """
        print(self.some_var)


class DummyControllerDevice(Device):
    USER_ACCESS = ["controller"]


class SynFlyer(Device, PositionerBase):
    def __init__(
        self,
        *,
        name,
        readback_func=None,
        value=0,
        delay=0,
        speed=1,
        update_frequency=2,
        precision=3,
        parent=None,
        labels=None,
        kind=None,
        device_manager=None,
        **kwargs,
    ):
        if readback_func is None:

            def readback_func(x):
                return x

        sentinel = object()
        loop = kwargs.pop("loop", sentinel)
        if loop is not sentinel:
            warnings.warn(
                f"{self.__class__} no longer takes a loop as input.  "
                "Your input will be ignored and may raise in the future",
                stacklevel=2,
            )
        self.sim_state = {}
        self._readback_func = readback_func
        self.delay = delay
        self.precision = precision
        self.tolerance = kwargs.pop("tolerance", 0.5)
        self.device_manager = device_manager

        # initialize values
        self.sim_state["readback"] = readback_func(value)
        self.sim_state["readback_ts"] = ttime.time()

        super().__init__(name=name, parent=parent, labels=labels, kind=kind, **kwargs)

    @property
    def hints(self):
        return {"fields": ["flyer_samx", "flyer_samy"]}

    def kickoff(self, metadata, num_pos, positions, exp_time: float = 0):
        positions = np.asarray(positions)

        def produce_data(device, metadata):
            buffer_time = 0.2
            elapsed_time = 0
            bundle = messages.BundleMessage()
            for ii in range(num_pos):
                bundle.append(
                    messages.DeviceMessage(
                        signals={
                            self.name: {
                                "flyer_samx": {"value": positions[ii, 0], "timestamp": 0},
                                "flyer_samy": {"value": positions[ii, 1], "timestamp": 0},
                            }
                        },
                        metadata={"pointID": ii, **metadata},
                    ).dumps()
                )
                ttime.sleep(exp_time)
                elapsed_time += exp_time
                if elapsed_time > buffer_time:
                    elapsed_time = 0
                    device.device_manager.producer.send(
                        MessageEndpoints.device_read(device.name), bundle.dumps()
                    )
                    bundle = messages.BundleMessage()
                    device.device_manager.producer.set_and_publish(
                        MessageEndpoints.device_status(device.name),
                        messages.DeviceStatusMessage(
                            device=device.name,
                            status=1,
                            metadata={"pointID": ii, **metadata},
                        ).dumps(),
                    )
            device.device_manager.producer.send(
                MessageEndpoints.device_read(device.name), bundle.dumps()
            )
            device.device_manager.producer.set_and_publish(
                MessageEndpoints.device_status(device.name),
                messages.DeviceStatusMessage(
                    device=device.name,
                    status=0,
                    metadata={"pointID": num_pos, **metadata},
                ).dumps(),
            )
            print("done")

        flyer = threading.Thread(target=produce_data, args=(self, metadata))
        flyer.start()


class SynController(OphydObject):
    def on(self):
        pass

    def off(self):
        pass


class SynFlyerLamNI(Device, PositionerBase):
    def __init__(
        self,
        *,
        name,
        readback_func=None,
        value=0,
        delay=0,
        speed=1,
        update_frequency=2,
        precision=3,
        parent=None,
        labels=None,
        kind=None,
        device_manager=None,
        **kwargs,
    ):
        if readback_func is None:

            def readback_func(x):
                return x

        sentinel = object()
        loop = kwargs.pop("loop", sentinel)
        if loop is not sentinel:
            warnings.warn(
                f"{self.__class__} no longer takes a loop as input.  "
                "Your input will be ignored and may raise in the future",
                stacklevel=2,
            )
        self.sim_state = {}
        self._readback_func = readback_func
        self.delay = delay
        self.precision = precision
        self.tolerance = kwargs.pop("tolerance", 0.5)
        self.device_manager = device_manager

        # initialize values
        self.sim_state["readback"] = readback_func(value)
        self.sim_state["readback_ts"] = ttime.time()

        super().__init__(name=name, parent=parent, labels=labels, kind=kind, **kwargs)
        self.controller = SynController(name="SynController")

    def kickoff(self, metadata, num_pos, positions, exp_time: float = 0):
        positions = np.asarray(positions)

        def produce_data(device, metadata):
            buffer_time = 0.2
            elapsed_time = 0
            bundle = messages.BundleMessage()
            for ii in range(num_pos):
                bundle.append(
                    messages.DeviceMessage(
                        signals={
                            "syn_flyer_lamni": {
                                "flyer_samx": {"value": positions[ii, 0], "timestamp": 0},
                                "flyer_samy": {"value": positions[ii, 1], "timestamp": 0},
                            }
                        },
                        metadata={"pointID": ii, **metadata},
                    ).dumps()
                )
                ttime.sleep(exp_time)
                elapsed_time += exp_time
                if elapsed_time > buffer_time:
                    elapsed_time = 0
                    device.device_manager.producer.send(
                        MessageEndpoints.device_read(device.name), bundle.dumps()
                    )
                    bundle = messages.BundleMessage()
                    device.device_manager.producer.set_and_publish(
                        MessageEndpoints.device_status(device.name),
                        messages.DeviceStatusMessage(
                            device=device.name,
                            status=1,
                            metadata={"pointID": ii, **metadata},
                        ).dumps(),
                    )
            device.device_manager.producer.send(
                MessageEndpoints.device_read(device.name), bundle.dumps()
            )
            device.device_manager.producer.set_and_publish(
                MessageEndpoints.device_status(device.name),
                messages.DeviceStatusMessage(
                    device=device.name,
                    status=0,
                    metadata={"pointID": num_pos, **metadata},
                ).dumps(),
            )
            print("done")

        flyer = threading.Thread(target=produce_data, args=(self, metadata))
        flyer.start()


class SynAxisOPAAS(Device, PositionerBase):
    """
    A synthetic settable Device mimic any 1D Axis (position, temperature).

    Parameters
    ----------
    name : string, keyword only
    readback_func : callable, optional
        When the Device is set to ``x``, its readback will be updated to
        ``f(x)``. This can be used to introduce random noise or a systematic
        offset.
        Expected signature: ``f(x) -> value``.
    value : object, optional
        The initial value. Default is 0.
    delay : number, optional
        Simulates how long it takes the device to "move". Default is 0 seconds.
    precision : integer, optional
        Digits of precision. Default is 3.
    parent : Device, optional
        Used internally if this Signal is made part of a larger Device.
    kind : a member the Kind IntEnum (or equivalent integer), optional
        Default is Kind.normal. See Kind for options.
    """

    USER_ACCESS = ["sim_state", "readback", "speed", "dummy_controller"]
    readback = Cpt(_ReadbackSignal, value=0, kind="hinted")
    setpoint = Cpt(_SetpointSignal, value=0, kind="normal")
    motor_is_moving = Cpt(_IsMovingSignal, value=0, kind="normal")

    velocity = Cpt(Signal, value=1, kind="config")
    acceleration = Cpt(Signal, value=1, kind="config")

    high_limit_travel = Cpt(Signal, value=0, kind="omitted")
    low_limit_travel = Cpt(Signal, value=0, kind="omitted")
    unused = Cpt(Signal, value=1, kind="omitted")

    SUB_READBACK = "readback"
    _default_sub = SUB_READBACK

    def __init__(
        self,
        *,
        name,
        readback_func=None,
        value=0,
        delay=0,
        speed=1,
        update_frequency=2,
        precision=3,
        parent=None,
        labels=None,
        kind=None,
        limits=None,
        **kwargs,
    ):
        if readback_func is None:

            def readback_func(x):
                return x

        sentinel = object()
        loop = kwargs.pop("loop", sentinel)
        if loop is not sentinel:
            warnings.warn(
                f"{self.__class__} no longer takes a loop as input.  "
                "Your input will be ignored and may raise in the future",
                stacklevel=2,
            )
        self.sim_state = {}
        self._readback_func = readback_func
        self.delay = delay
        self.precision = precision
        self.speed = speed
        self.update_frequency = update_frequency
        self.tolerance = kwargs.pop("tolerance", 0.05)
        self._stopped = False

        self.dummy_controller = DummyController()

        # initialize values
        self.sim_state["setpoint"] = value
        self.sim_state["setpoint_ts"] = ttime.time()
        self.sim_state["readback"] = readback_func(value)
        self.sim_state["readback_ts"] = ttime.time()
        self.sim_state["is_moving"] = 0
        self.sim_state["is_moving_ts"] = ttime.time()

        super().__init__(name=name, parent=parent, labels=labels, kind=kind, **kwargs)
        self.readback.name = self.name
        if limits is not None:
            assert len(limits) == 2
            self.low_limit_travel.put(limits[0])
            self.high_limit_travel.put(limits[1])

    @property
    def limits(self):
        return (self.low_limit_travel.get(), self.high_limit_travel.get())

    @property
    def low_limit(self):
        return self.limits[0]

    @property
    def high_limit(self):
        return self.limits[1]

    def check_value(self, pos):
        """Check that the position is within the soft limits"""
        low_limit, high_limit = self.limits

        if low_limit < high_limit and not (low_limit <= pos <= high_limit):
            raise LimitError(f"position={pos} not within limits {self.limits}")

    def set(self, value):
        self._stopped = False
        self.check_value(value)
        old_setpoint = self.sim_state["setpoint"]
        self.sim_state["is_moving"] = 1
        self.motor_is_moving._run_subs(
            sub_type=self.motor_is_moving.SUB_VALUE,
            old_value=0,
            value=1,
            timestamp=self.sim_state["is_moving_ts"],
        )
        self.sim_state["setpoint"] = value
        self.sim_state["setpoint_ts"] = ttime.time()
        self.setpoint._run_subs(
            sub_type=self.setpoint.SUB_VALUE,
            old_value=old_setpoint,
            value=self.sim_state["setpoint"],
            timestamp=self.sim_state["setpoint_ts"],
        )

        def update_state(val):
            if self._stopped:
                raise DeviceStop
            old_readback = self.sim_state["readback"]
            self.sim_state["readback"] = val
            self.sim_state["readback_ts"] = ttime.time()
            self.readback._run_subs(
                sub_type=self.readback.SUB_VALUE,
                old_value=old_readback,
                value=self.sim_state["readback"],
                timestamp=self.sim_state["readback_ts"],
            )
            self._run_subs(
                sub_type=self.SUB_READBACK,
                old_value=old_readback,
                value=self.sim_state["readback"],
                timestamp=self.sim_state["readback_ts"],
            )

        st = DeviceStatus(device=self)
        if self.delay:

            def move_and_finish():
                success = True
                try:
                    move_val = self.sim_state["setpoint"] + self.tolerance * np.random.uniform(
                        -1, 1
                    )
                    updates = np.ceil(
                        np.abs(old_setpoint - move_val) / self.speed * self.update_frequency
                    )
                    for ii in np.linspace(old_setpoint, move_val, int(updates)):
                        ttime.sleep(1 / self.update_frequency)
                        update_state(ii)
                    update_state(move_val)
                    self.sim_state["is_moving"] = 0
                    self.sim_state["is_moving_ts"] = ttime.time()
                    self.motor_is_moving._run_subs(
                        sub_type=self.motor_is_moving.SUB_VALUE,
                        old_value=1,
                        value=0,
                        timestamp=self.sim_state["is_moving_ts"],
                    )
                except DeviceStop:
                    success = False
                finally:
                    self._stopped = False
                self._done_moving(success=success)
                st.set_finished()

            threading.Thread(target=move_and_finish, daemon=True).start()
            # raise TimeoutError

        else:
            update_state(value)
            self._done_moving()
            st.set_finished()
        return st

    def stop(self, *, success=False):
        super().stop(success=success)
        self._stopped = True

    @property
    def position(self):
        return self.readback.get()

    def egu(self):
        return "mm"


class SynGaussBEC(Device):
    """
    Evaluate a point on a Gaussian based on the value of a motor.

    Parameters
    ----------
    name : string
    motor : Device
    motor_field : string
    center : number
        center of peak
    Imax : number
        max intensity of peak
    sigma : number, optional
        Default is 1.
    noise : {'poisson', 'uniform', None}, optional
        Add noise to the gaussian peak.
    noise_multiplier : float, optional
        Only relevant for 'uniform' noise. Multiply the random amount of
        noise by 'noise_multiplier'
    random_state : numpy random state object, optional
        np.random.RandomState(0), to generate random number with given seed

    Example
    -------
    motor = SynAxis(name='motor')
    det = SynGauss('det', motor, 'motor', center=0, Imax=1, sigma=1)
    """

    val = Cpt(_ReadbackSignalCompute, value=0, kind="hinted")
    Imax = Cpt(Signal, value=10, kind="config")
    center = Cpt(Signal, value=0, kind="config")
    sigma = Cpt(Signal, value=1, kind="config")
    motor = Cpt(Signal, value="samx", kind="config")
    noise = Cpt(
        EnumSignal,
        value="none",
        kind="config",
        enum_strings=("none", "poisson", "uniform"),
    )
    noise_multiplier = Cpt(Signal, value=1, kind="config")

    def __init__(self, name, *, device_manager=None, random_state=None, **kwargs):
        self.device_manager = device_manager
        set_later = {}
        for k in ("sigma", "noise", "noise_multiplier"):
            v = kwargs.pop(k, None)
            if v is not None:
                set_later[k] = v
        super().__init__(name=name, **kwargs)

        self.random_state = random_state or np.random
        self.val.name = self.name
        self.precision = 3
        self.sim_state = {"readback": 0, "readback_ts": ttime.time()}
        for k, v in set_later.items():
            getattr(self, k).put(v)

    def _compute(self):
        try:
            m = self.device_manager.devices[self.motor.get()].obj.read()[self.motor.get()]["value"]
            # we need to do this one at a time because
            #   - self.read() may be screwed with by the user
            #   - self.get() would cause infinite recursion
            Imax = self.Imax.get()
            center = self.center.get()
            sigma = self.sigma.get()
            noise = self.noise.get()
            noise_multiplier = self.noise_multiplier.get()
            v = Imax * np.exp(-((m - center) ** 2) / (2 * sigma**2))
            if noise == "poisson":
                v = int(self.random_state.poisson(np.round(v), 1))
            elif noise == "uniform":
                v += self.random_state.uniform(-1, 1) * noise_multiplier
            return v
        except Exception:
            return 0

    def get(self, *args, **kwargs):
        self.sim_state["readback"] = self._compute()
        self.sim_state["readback_ts"] = ttime.time()
        return self.val.get()


class SynDynamicComponents(Device):
    messages = Dcpt({f"message{i}": (SynSignal, None, {"name": f"msg{i}"}) for i in range(1, 6)})


class SynDeviceSubOPAAS(Device):
    zsub = Cpt(SynAxisOPAAS, name="zsub")


class SynDeviceOPAAS(Device):
    x = Cpt(SynAxisOPAAS, name="x")
    y = Cpt(SynAxisOPAAS, name="y")
    z = Cpt(SynDeviceSubOPAAS, name="z")


if __name__ == "__main__":
    det = SynSLSDetector(name="moench")
    det.trigger()



import enum
import threading
import time
import numpy as np
import os

from typing import Any, List

from ophyd import EpicsSignal, EpicsSignalRO, EpicsSignalWithRBV
from ophyd import Device
from ophyd import ADComponent as ADCpt

from std_daq_client import StdDaqClient

from bec_lib import threadlocked
from bec_lib.logger import bec_logger
from bec_lib import messages
from bec_lib.endpoints import MessageEndpoints

from ophyd_devices.epics.devices.psi_detector_base import PSIDetectorBase, CustomDetectorMixin

logger = bec_logger.logger


class EigerError(Exception):
    """Base class for exceptions in this module."""


class EigerTimeoutError(EigerError):
    """Raised when the Eiger does not respond in time."""


class Eiger9MSetup(CustomDetectorMixin):
    """Eiger setup class

    Parent class: CustomDetectorMixin

    """

    def __init__(self, *args, parent: Device = None, **kwargs) -> None:
        super().__init__(*args, parent=parent, **kwargs)
        self.std_rest_server_url = (
            kwargs["file_writer_url"] if "file_writer_url" in kwargs else "http://xbl-daq-29:5000"
        )
        self.std_client = None
        self._lock = threading.RLock()

    def initialize_default_parameter(self) -> None:
        """Set default parameters for Eiger9M detector"""
        self.update_readout_time()

    def update_readout_time(self) -> None:
        """Set readout time for Eiger9M detector"""
        readout_time = (
            self.parent.scaninfo.readout_time
            if hasattr(self.parent.scaninfo, "readout_time")
            else self.parent.MIN_READOUT
        )
        self.parent.readout_time = max(readout_time, self.parent.MIN_READOUT)

    def initialize_detector(self) -> None:
        """Initialize detector"""
        # Stops the detector
        self.stop_detector()
        # Sets the trigger source to GATING
        self.parent.set_trigger(TriggerSource.GATING)

    def initialize_detector_backend(self) -> None:
        """Initialize detector backend"""

        # Std client
        self.std_client = StdDaqClient(url_base=self.std_rest_server_url)

        # Stop writer
        self.std_client.stop_writer()

        # Change e-account
        eacc = self.parent.scaninfo.username
        self.update_std_cfg("writer_user_id", int(eacc.strip(" e")))

        signal_conditions = [
            (
                lambda: self.std_client.get_status()["state"],
                "READY",
            ),
        ]
        if not self.wait_for_signals(
            signal_conditions=signal_conditions,
            timeout=self.parent.timeout,
            all_signals=True,
        ):
            raise EigerTimeoutError(
                f"Std client not in READY state, returns: {self.std_client.get_status()}"
            )

    def update_std_cfg(self, cfg_key: str, value: Any) -> None:
        """
        Update std_daq config

        Checks that the new value matches the type of the former entry.

        Args:
            cfg_key (str)   : config key of value to be updated
            value (Any)     : value to be updated for the specified key

        Raises:
            Raises EigerError if the key was not in the config before and if the new value does not match the type of the old value

        """

        # Load config from client and check old value
        cfg = self.std_client.get_config()
        old_value = cfg.get(cfg_key)
        if old_value is None:
            raise EigerError(
                f"Tried to change entry for key {cfg_key} in std_config that does not exist"
            )
        if not isinstance(value, type(old_value)):
            raise EigerError(
                f"Type of new value {type(value)}:{value} does not match old value"
                f" {type(old_value)}:{old_value}"
            )

        # Update config with new value and send back to client
        cfg.update({cfg_key: value})
        logger.debug(cfg)
        self.std_client.set_config(cfg)
        logger.debug(f"Updated std_daq config for key {cfg_key} from {old_value} to {value}")

    def stop_detector(self) -> None:
        """Stop the detector"""

        # Stop detector
        self.parent.cam.acquire.put(0)

        # Check if detector returned in idle state
        signal_conditions = [
            (
                lambda: self.parent.cam.detector_state.read()[self.parent.cam.detector_state.name][
                    "value"
                ],
                DetectorState.IDLE,
            )
        ]

        if not self.wait_for_signals(
            signal_conditions=signal_conditions,
            timeout=self.parent.timeout - self.parent.timeout // 2,
            check_stopped=True,
            all_signals=False,
        ):
            # Retry stop detector and wait for remaining time
            self.parent.cam.acquire.put(0)
            if not self.wait_for_signals(
                signal_conditions=signal_conditions,
                timeout=self.parent.timeout - self.parent.timeout // 2,
                check_stopped=True,
                all_signals=False,
            ):
                raise EigerTimeoutError(
                    f"Failed to stop detector, detector state {signal_conditions[0][0]}"
                )

    def stop_detector_backend(self) -> None:
        """Close file writer"""
        self.std_client.stop_writer()

    def prepare_detector(self) -> None:
        """Prepare detector for scan"""
        self.set_detector_threshold()
        self.set_acquisition_params()
        self.parent.set_trigger(TriggerSource.GATING)

    def set_detector_threshold(self) -> None:
        """
        Set the detector threshold

        The function sets the detector threshold automatically to 1/2 of the beam energy.
        """

        # get current beam energy from device manageer
        mokev = self.parent.device_manager.devices.mokev.obj.read()[
            self.parent.device_manager.devices.mokev.name
        ]["value"]
        factor = 1

        # Check if energies are eV or keV, assume keV as the default
        unit = getattr(self.parent.cam.threshold_energy, "units", None)
        if unit is not None and unit == "eV":
            factor = 1000

        # set energy on detector
        setpoint = int(mokev * factor)
        energy = self.parent.cam.beam_energy.read()[self.parent.cam.beam_energy.name]["value"]
        if setpoint != energy:
            self.parent.cam.beam_energy.set(setpoint)

        # set threshold on detector
        threshold = self.parent.cam.threshold_energy.read()[self.parent.cam.threshold_energy.name][
            "value"
        ]
        if not np.isclose(setpoint / 2, threshold, rtol=0.05):
            self.parent.cam.threshold_energy.set(setpoint / 2)

    def set_acquisition_params(self) -> None:
        """Set acquisition parameters for the detector"""

        # Set number of images and frames (frames is for internal burst of detector)
        self.parent.cam.num_images.put(
            int(self.parent.scaninfo.num_points * self.parent.scaninfo.frames_per_trigger)
        )
        self.parent.cam.num_frames.put(1)

        # Update the readout time of the detector
        self.update_readout_time()

    def prepare_data_backend(self) -> None:
        """Prepare the data backend for the scan"""
        self.parent.filepath = self.parent.filewriter.compile_full_filename(
            self.parent.scaninfo.scan_number, f"{self.parent.name}.h5", 1000, 5, True
        )
        self.filepath_exists(self.parent.filepath)
        self.stop_detector_backend()
        try:
            self.std_client.start_writer_async(
                {
                    "output_file": self.parent.filepath,
                    "n_images": int(
                        self.parent.scaninfo.num_points * self.parent.scaninfo.frames_per_trigger
                    ),
                }
            )
        except Exception as exc:
            time.sleep(5)
            if self.std_client.get_status()["state"] == "READY":
                raise EigerTimeoutError(f"Timeout of start_writer_async with {exc}") from exc

        # Check status of std_daq
        signal_conditions = [
            (lambda: self.std_client.get_status()["acquisition"]["state"], "WAITING_IMAGES")
        ]
        if not self.wait_for_signals(
            signal_conditions=signal_conditions,
            timeout=self.parent.timeout,
            check_stopped=False,
            all_signals=True,
        ):
            raise EigerTimeoutError(
                "Timeout of 5s reached for std_daq start_writer_async with std_daq client status"
                f" {self.std_client.get_status()}"
            )

    def filepath_exists(self, filepath: str) -> None:
        """Check if filepath exists"""
        signal_conditions = [(lambda: os.path.exists(os.path.dirname(filepath)), True)]
        if not self.wait_for_signals(
            signal_conditions=signal_conditions,
            timeout=self.parent.timeout,
            check_stopped=False,
            all_signals=True,
        ):
            raise EigerError(f"Timeout of 3s reached for filepath {filepath}")

    def arm_acquisition(self) -> None:
        """Arm Eiger detector for acquisition"""
        self.parent.cam.acquire.put(1)
        signal_conditions = [
            (
                lambda: self.parent.cam.detector_state.read()[self.parent.cam.detector_state.name][
                    "value"
                ],
                DetectorState.RUNNING,
            )
        ]
        if not self.wait_for_signals(
            signal_conditions=signal_conditions,
            timeout=self.parent.timeout,
            check_stopped=True,
            all_signals=False,
        ):
            raise EigerTimeoutError(
                f"Failed to arm the acquisition. Detector state {signal_conditions[0][0]}"
            )

    def check_scanID(self) -> None:
        """Checks if scanID has changed and stops the scan if it has"""
        old_scanID = self.parent.scaninfo.scanID
        self.parent.scaninfo.load_scan_metadata()
        if self.parent.scaninfo.scanID != old_scanID:
            self.parent.stopped = True

    def publish_file_location(self, done: bool = False, successful: bool = None) -> None:
        """
        Publish the filepath to REDIS.

        We publish two events here:
        - file_event: event for the filewriter
        - public_file: event for any secondary service (e.g. radial integ code)

        Args:
            done (bool): True if scan is finished
            successful (bool): True if scan was successful
        """
        pipe = self.parent.producer.pipeline()
        if successful is None:
            msg = messages.FileMessage(file_path=self.parent.filepath, done=done)
        else:
            msg = messages.FileMessage(
                file_path=self.parent.filepath, done=done, successful=successful
            )
        self.parent.producer.set_and_publish(
            MessageEndpoints.public_file(self.parent.scaninfo.scanID, self.parent.name),
            msg.dumps(),
            pipe=pipe,
        )
        self.parent.producer.set_and_publish(
            MessageEndpoints.file_event(self.parent.name), msg.dumps(), pipe=pipe
        )
        pipe.execute()

    @threadlocked
    def finished(self):
        """Check if acquisition is finished."""
        signal_conditions = [
            (
                lambda: self.parent.cam.acquire.read()[self.parent.cam.acquire.name]["value"],
                DetectorState.IDLE,
            ),
            (lambda: self.std_client.get_status()["acquisition"]["state"], "FINISHED"),
            (
                lambda: self.std_client.get_status()["acquisition"]["stats"]["n_write_completed"],
                int(self.parent.scaninfo.num_points * self.parent.scaninfo.frames_per_trigger),
            ),
        ]
        if not self.wait_for_signals(
            signal_conditions=signal_conditions,
            timeout=self.parent.timeout,
            check_stopped=True,
            all_signals=True,
        ):
            raise EigerTimeoutError(
                f"Reached timeout with detector state {signal_conditions[0][0]}, std_daq state"
                f" {signal_conditions[1][0]} and received frames of {signal_conditions[2][0]} for"
                " the file writer"
            )
        self.stop_detector()
        self.stop_detector_backend()


class SLSDetectorCam(Device):
    """
    SLS Detector Camera - Eiger9M

    Base class to map EPICS PVs to ophyd signals.
    """

    threshold_energy = ADCpt(EpicsSignalWithRBV, "ThresholdEnergy")
    beam_energy = ADCpt(EpicsSignalWithRBV, "BeamEnergy")
    bit_depth = ADCpt(EpicsSignalWithRBV, "BitDepth")
    num_images = ADCpt(EpicsSignalWithRBV, "NumCycles")
    num_frames = ADCpt(EpicsSignalWithRBV, "NumFrames")
    trigger_mode = ADCpt(EpicsSignalWithRBV, "TimingMode")
    trigger_software = ADCpt(EpicsSignal, "TriggerSoftware")
    acquire = ADCpt(EpicsSignal, "Acquire")
    detector_state = ADCpt(EpicsSignalRO, "DetectorState_RBV")


class TriggerSource(enum.IntEnum):
    """Trigger signals for Eiger9M detector"""

    AUTO = 0
    TRIGGER = 1
    GATING = 2
    BURST_TRIGGER = 3


class DetectorState(enum.IntEnum):
    """Detector states for Eiger9M detector"""

    IDLE = 0
    ERROR = 1
    WAITING = 2
    FINISHED = 3
    TRANSMITTING = 4
    RUNNING = 5
    STOPPED = 6
    STILL_WAITING = 7
    INITIALIZING = 8
    DISCONNECTED = 9
    ABORTED = 10


class Eiger9McSAXS(PSIDetectorBase):
    """
    Eiger9M detector for CSAXS

    Parent class: PSIDetectorBase

    class attributes:
        custom_prepare_cls (FalconSetup)        : Custom detector setup class for cSAXS,
                                                  inherits from CustomDetectorMixin
        PSIDetectorBase.set_min_readout (float) : Minimum readout time for the detector
        Various EpicsPVs for controlling the detector
    """

    # Specify which functions are revealed to the user in BEC client
    USER_ACCESS = [
        "describe",
    ]

    # specify Setup class
    custom_prepare_cls = Eiger9MSetup
    # specify minimum readout time for detector
    MIN_READOUT = 3e-3
    # specify class attributes
    cam = ADCpt(SLSDetectorCam, "cam1:")

    def set_trigger(self, trigger_source: TriggerSource) -> None:
        """Set trigger source for the detector.
        Check the TriggerSource enum for possible values

        Args:
            trigger_source (TriggerSource): Trigger source for the detector

        """
        value = trigger_source
        self.cam.trigger_mode.put(value)

    def stage(self) -> List[object]:
        """
        Add functionality to stage, and arm the detector

        Additional call to:
        - custom_prepare.arm_acquisition()
        """
        rtr = super().stage()
        self.custom_prepare.arm_acquisition()
        return rtr


if __name__ == "__main__":
    eiger = Eiger9McSAXS(name="eiger", prefix="X12SA-ES-EIGER9M:", sim_mode=True)



from .DelayGeneratorDG645 import DelayGeneratorDG645
from .slits import SlitH, SlitV
from .XbpmBase import XbpmBase, XbpmCsaxsOp
from .SpmBase import SpmBase
from .InsertionDevice import InsertionDevice
from .specMotors import (
    Bpm4i,
    EnergyKev,
    GirderMotorPITCH,
    GirderMotorROLL,
    GirderMotorX1,
    GirderMotorY1,
    GirderMotorYAW,
    MonoTheta1,
    MonoTheta2,
    PmDetectorRotation,
    PmMonoBender,
)

# Standard ophyd classes
from ophyd import EpicsSignal, EpicsSignalRO, EpicsMotor
from ophyd.sim import SynAxis, SynSignal, SynPeriodicSignal
from ophyd.quadem import QuadEM

# cSAXS
from .epics_motor_ex import EpicsMotorEx
from .mcs_csaxs import McsCsaxs
from .psi_detector_base import PSIDetectorBase, CustomDetectorMixin
from .eiger9m_csaxs import Eiger9McSAXS
from .pilatus_csaxs import PilatuscSAXS
from .falcon_csaxs import FalconcSAXS
from .DelayGeneratorDG645 import DelayGeneratorDG645

# from .psi_detector_base import PSIDetectorBase, CustomDetectorMixin



from .eiger1p5m_csaxs.eiger1p5m import Eiger1p5MDetector
from .epics import *
from .galil.galil_ophyd import GalilMotor
from .galil.sgalil_ophyd import SGalilMotor
from .npoint.npoint import NPointAxis
from .rt_lamni import RtLamniMotor
from .sim.sim import (
    SynAxisMonitor,
    SynAxisOPAAS,
    SynDeviceOPAAS,
    SynFlyer,
    SynSignalRO,
    SynSLSDetector,
)
from .sls_devices.sls_devices import SLSInfo, SLSOperatorMessages
from .smaract.smaract_ophyd import SmaractMotor



import functools
import threading
import time
from typing import List

import numpy as np
from bec_lib import bec_logger
from ophyd import Component as Cpt, DeviceStatus
from ophyd import Device, PositionerBase, Signal
from ophyd.status import wait as status_wait
from ophyd.utils import LimitError, ReadOnlyError
from ophyd_devices.utils.controller import Controller, threadlocked
from ophyd_devices.utils.socket import SocketIO, SocketSignal, raise_if_disconnected
from prettytable import PrettyTable

logger = bec_logger.logger


class GalilCommunicationError(Exception):
    pass


class GalilError(Exception):
    pass


class BECConfigError(Exception):
    pass


def retry_once(fcn):
    """Decorator to rerun a function in case a Galil communication error was raised. This may happen if the buffer was not empty."""

    @functools.wraps(fcn)
    def wrapper(self, *args, **kwargs):
        try:
            val = fcn(self, *args, **kwargs)
        except (GalilCommunicationError, GalilError):
            val = fcn(self, *args, **kwargs)
        return val

    return wrapper


class GalilController(Controller):
    USER_ACCESS = [
        "describe",
        "show_running_threads",
        "galil_show_all",
        "socket_put_and_receive",
        "socket_put_confirmed",
        "sgalil_reference",
        "fly_grid_scan",
        "read_encoder_position",
    ]

    def __init__(
        self,
        *,
        name="GalilController",
        kind=None,
        parent=None,
        socket=None,
        attr_name="",
        labels=None,
    ):
        if not hasattr(self, "_initialized") or not self._initialized:
            self._galil_axis_per_controller = 8
            self._axis = [None for axis_num in range(self._galil_axis_per_controller)]
            super().__init__(
                name=name,
                socket=socket,
                attr_name=attr_name,
                parent=parent,
                labels=labels,
                kind=kind,
            )

    def on(self, controller_num=0) -> None:
        """Open a new socket connection to the controller"""
        if not self.connected:
            self.sock.open()
            self.connected = True
        else:
            logger.info("The connection has already been established.")
            # warnings.warn(f"The connection has already been established.", stacklevel=2)

    def off(self) -> None:
        """Close the socket connection to the controller"""
        if self.connected:
            self.sock.close()
            self.connected = False
        else:
            logger.info("The connection is already closed.")

    def set_axis(self, axis: Device, axis_nr: int) -> None:
        """Assign an axis to a device instance.

        Args:
            axis (Device): Device instance (e.g. GalilMotor)
            axis_nr (int): Controller axis number

        """
        self._axis[axis_nr] = axis

    @threadlocked
    def socket_put(self, val: str) -> None:
        time.sleep(0.01)
        self.sock.put(f"{val}\r".encode())

    @threadlocked
    def socket_get(self) -> str:
        time.sleep(0.01)
        return self.sock.receive().decode()

    @retry_once
    @threadlocked
    def socket_put_and_receive(self, val: str, remove_trailing_chars=True) -> str:
        self.socket_put(val)
        if remove_trailing_chars:
            return self._remove_trailing_characters(self.sock.receive().decode())
        return self.socket_get()

    @retry_once
    def socket_put_confirmed(self, val: str) -> None:
        """Send message to controller and ensure that it is received by checking that the socket receives a colon.

        Args:
            val (str): Message that should be sent to the socket

        Raises:
            GalilCommunicationError: Raised if the return value is not a colon.

        """
        return_val = self.socket_put_and_receive(val)
        if return_val != ":":
            raise GalilCommunicationError(
                f"Expected return value of ':' but instead received {return_val}"
            )

    def is_axis_moving(self, axis_Id, axis_Id_numeric) -> bool:
        is_moving = bool(float(self.socket_put_and_receive(f"MG_BG{axis_Id}")) != 0)
        # backlash_is_active = bool(float(self.socket_put_and_receive(f"MGbcklact[axis]")) != 0)
        return bool(is_moving)  # bool(is_moving or backlash_is_active)

    def is_thread_active(self, thread_id: int) -> bool:
        val = float(self.socket_put_and_receive(f"MG_XQ{thread_id}"))
        if val == -1:
            return False
        return True

    def _remove_trailing_characters(self, var) -> str:
        if len(var) > 1:
            return var.split("\r\n")[0]
        return var

    def stop_all_axes(self) -> str:
        # return self.socket_put_and_receive(f"XQ#STOP,1")
        # Command stops all threads and motors!
        self.socket_put_and_receive(f"CB8")
        return self.socket_put_and_receive(f"ST")

    def axis_is_referenced(self) -> bool:
        return bool(float(self.socket_put_and_receive(f"MG allaxref").strip()))

    def show_running_threads(self) -> None:
        t = PrettyTable()
        t.title = f"Threads on {self.sock.host}:{self.sock.port}"
        t.field_names = [str(ax) for ax in range(self._galil_axis_per_controller)]
        t.add_row(
            [
                "active" if self.is_thread_active(t) else "inactive"
                for t in range(self._galil_axis_per_controller)
            ]
        )
        print(t)

    def is_motor_on(self, axis_Id) -> bool:
        return not bool(float(self.socket_put_and_receive(f"MG _MO{axis_Id}").strip()))

    def get_motor_limit_switch(self, axis_Id) -> list:
        # SGalil specific
        if axis_Id == "C":
            ret = self.socket_put_and_receive(f"MG _LF{axis_Id}, _LR{axis_Id}")
            high, low = ret.strip().split(" ")
        elif axis_Id == "E":
            ret = self.socket_put_and_receive(f"MG _LF{'F'}, _LR{'F'}")
            high, low = ret.strip().split(" ")
        return [not bool(float(low)), not bool(float(high))]

    def describe(self) -> None:
        t = PrettyTable()
        t.title = f"{self.__class__.__name__} on {self.sock.host}:{self.sock.port}"
        t.field_names = [
            "Axis",
            "Name",
            "Connected",
            "Referenced",
            "Motor On",
            "Limits",
            "Position",
        ]
        for ax in range(self._galil_axis_per_controller):
            axis = self._axis[ax]
            if axis is not None:
                t.add_row(
                    [
                        f"{axis.axis_Id_numeric}/{axis.axis_Id}",
                        axis.name,
                        axis.connected,
                        self.axis_is_referenced(),
                        self.is_motor_on(axis.axis_Id),
                        self.get_motor_limit_switch(axis.axis_Id),
                        axis.readback.read().get(axis.name).get("value"),
                    ]
                )
            else:
                t.add_row([None for t in t.field_names])
        print(t)

        self.show_running_threads()

    def galil_show_all(self) -> None:
        for controller in self._controller_instances.values():
            if isinstance(controller, GalilController):
                controller.describe()

    def sgalil_reference(self) -> None:
        """Reference all axes of the controller"""
        if self.axis_is_referenced():
            print("All axes are already referenced.\n")
            return
        # Make sure no axes are moving, is this necessary?
        self.stop_all_axes()
        self.socket_put_and_receive(f"XQ#FINDREF")
        print("Referencing. Please wait, timeout after 100s...\n")

        timeout = time.time() + 100
        while not self.axis_is_referenced():
            if time.time() > timeout:
                print("Abort reference sequence, timeout reached\n")
                break
            time.sleep(0.5)

    # @threadlocked
    def fly_grid_scan(
        self,
        status: DeviceStatus,
        start_y: float,
        end_y: float,
        interval_y: int,
        start_x: float,
        end_x: float,
        interval_x: int,
        exp_time: float,
        readout_time: float,
        **kwargs,
    ) -> tuple:
        """_summary_

        Args:
            start_y (float): start position of y axis (fast axis)
            end_y (float): end position of y axis (fast axis)
            interval_y (int): number of points in y axis
            start_x (float): start position of x axis (slow axis)
            end_x (float): end position of x axis (slow axis)
            interval_x (int): number of points in x axis
            exp_time (float): exposure time in seconds
            readout_time (float): readout time in seconds, minimum of .5e-3s (0.5ms)

        Raises:

            LimitError: Raised if any position of motion is outside of the limits
            LimitError: Raised if the speed is above 2mm/s or below 0.02mm/s

        """
        #
        if not self.axis_is_referenced():
            raise GalilError("Axis are not referenced")
        sign_y = self._axis[ord("c") - 97].sign
        sign_x = self._axis[ord("e") - 97].sign
        # Check limits
        # TODO check sign of stage, or not necessary
        check_values = [start_y, end_y, start_x, end_x]
        for val in check_values:
            self.check_value(val)

        start_x *= sign_x
        end_x *= sign_x
        start_y *= sign_y
        end_y *= sign_y

        speed = np.abs(end_y - start_y) / (
            (interval_y) * exp_time + (interval_y - 1) * readout_time
        )
        if speed > 2.00 or speed < 0.02:
            raise LimitError(
                f"Speed of {speed:.03f}mm/s is outside of acceptable range of 0.02 to 2 mm/s"
            )

        gridmax = int(interval_x - 1)
        step_grid = (end_x - start_x) / interval_x
        n_samples = int(interval_y * interval_x)

        # Hard coded to maximum offset of 0.1mm to avoid long motions.
        self.socket_put_and_receive(f"off={(0):f}")
        self.socket_put_and_receive(f"a_start={start_y:.04f};a_end={end_y:.04f};speed={speed:.04f}")
        self.socket_put_and_receive(
            f"b_start={start_x:.04f};gridmax={gridmax:d};b_step={step_grid:.04f}"
        )
        self.socket_put_and_receive(f"nums={n_samples}")
        self.socket_put_and_receive("XQ#SAMPLE")
        # sleep 50ms to avoid controller running into
        time.sleep(0.1)
        self.socket_put_and_receive("XQ#SCANG")
        # self._block_while_active(3)
        # time.sleep(0.1)
        threading.Thread(target=self._block_while_active, args=(3, status), daemon=True).start()
        # self._while_in_motion(3, n_samples)

    def _block_while_active(self, thread_id: int, status) -> None:
        while self.is_thread_active(thread_id):
            time.sleep(1)
        time.sleep(1)
        while self.is_thread_active(thread_id):
            time.sleep(1)
        status.set_finished()

    # TODO this is for reading out positions, readout is limited by stage triggering
    def _while_in_motion(self, thread_id: int, n_samples: int) -> tuple:
        last_readout = 0
        val_axis2 = []  # y axis
        val_axis4 = []  # x axis
        while self.is_thread_active(thread_id):
            posct = int(self.socket_put_and_receive(f"MGposct").strip().split(".")[0])
            logger.info(f"SGalil is scanning - latest enconder position {posct+1} from {n_samples}")
            time.sleep(1)
            if posct > last_readout:
                positions = self.read_encoder_position(last_readout, posct)
                val_axis4.extend(positions[0])
                val_axis2.extend(positions[1])
                last_readout = posct + 1
            logger.info(len(val_axis2))
            time.sleep(1)
        # Readout of last positions after scan finished
        posct = int(self.socket_put_and_receive(f"MGposct").strip().split(".")[0])
        logger.info(f"SGalil is scanning - latest enconder position {posct} from {n_samples}")
        if posct > last_readout:
            positions = self.read_encoder_position(last_readout, posct)
            val_axis4.extend(positions[0])
            val_axis2.extend(positions[1])

        return val_axis4, val_axis2

    def read_encoder_position(self, fromval: int, toval: int) -> tuple:
        val_axis2 = []  # y axis
        val_axis4 = []  # x axis
        for ii in range(fromval, toval + 1):
            rts = self.socket_put_and_receive(f"MGaposavg[{ii%2000}]*10,cposavg[{ii%2000}]*10")
            if rts == ":":
                val_axis4.append(rts)
                val_axis2.append(rts)
                continue

            val_axis4.append(float(rts.strip().split(" ")[0]) / 100000)
            val_axis2.append(float(rts.strip().split(" ")[1]) / 100000)
        return val_axis4, val_axis2


class GalilSignalBase(SocketSignal):
    def __init__(self, signal_name, **kwargs):
        self.signal_name = signal_name
        super().__init__(**kwargs)
        self.controller = self.parent.controller
        self.sock = self.parent.controller.sock


class GalilSignalRO(GalilSignalBase):
    def __init__(self, signal_name, **kwargs):
        super().__init__(signal_name, **kwargs)
        self._metadata["write_access"] = False

    def _socket_set(self, val):
        raise ReadOnlyError("Read-only signals cannot be set")


class GalilReadbackSignal(GalilSignalRO):
    @retry_once
    @threadlocked
    def _socket_get(self) -> float:
        """Get command for the readback signal

        Returns:
            float: Readback value after adjusting for sign and motor resolution.
        """
        if self.parent.axis_Id_numeric == 2:
            current_pos = float(
                self.controller.socket_put_and_receive(f"MG _TP{self.parent.axis_Id}/mm")
            )
        elif self.parent.axis_Id_numeric == 4:
            # hardware controller readback from axis 4 is on axis 0, A instead of E
            current_pos = float(self.controller.socket_put_and_receive(f"MG _TP{'A'}/mm"))
        current_pos *= self.parent.sign
        return current_pos

    def read(self):
        self._metadata["timestamp"] = time.time()
        val = super().read()
        return val


class GalilSetpointSignal(GalilSignalBase):
    setpoint = 0

    def _socket_get(self) -> float:
        """Get command for receiving the setpoint / target value.
        The value is not pulled from the controller but instead just the last setpoint used.

        Returns:
            float: setpoint / target value
        """
        return self.setpoint * self.parent.sign

    @retry_once
    @threadlocked
    def _socket_set(self, val: float) -> None:
        """Set a new target value / setpoint value. Before submission, the target value is adjusted for the axis' sign.
        Furthermore, it is ensured that all axes are referenced before a new setpoint is submitted.

        Args:
            val (float): Target value / setpoint value

        Raises:
            GalilError: Raised if not all axes are referenced.

        """
        target_val = val * self.parent.sign
        self.setpoint = target_val
        axes_referenced = self.controller.axis_is_referenced()
        if not axes_referenced:
            raise GalilError(
                "Not all axes are referenced. Please use controller.sgalil_reference(). BE AWARE that axes start moving, potentially beyond limits, make sure full range of motion is safe"
            )
        while self.controller.is_thread_active(0):
            time.sleep(0.1)

        if self.parent.axis_Id_numeric == 2:
            self.controller.socket_put_confirmed(f"PA{self.parent.axis_Id}={target_val:.4f}*mm")
            self.controller.socket_put_and_receive(f"BG{self.parent.axis_Id}")
        elif self.parent.axis_Id_numeric == 4:
            self.controller.socket_put_confirmed(f"targ{self.parent.axis_Id}={target_val:.4f}")
            self.controller.socket_put_and_receive(f"XQ#POSE,{self.parent.axis_Id_numeric}")
        while self.controller.is_thread_active(0):
            time.sleep(0.005)


class GalilMotorIsMoving(GalilSignalRO):
    @threadlocked
    def _socket_get(self):
        if self.parent.axis_Id_numeric == 2:
            ret = self.controller.is_axis_moving(self.parent.axis_Id, self.parent.axis_Id_numeric)
            return ret
        if self.parent.axis_Id_numeric == 4:
            # Motion signal from axis 4 is mapped to axis 5
            ret = self.controller.is_axis_moving("F", 5)
            return ret or self.controller.is_thread_active(4)

    def get(self):
        val = super().get()
        if val is not None:
            self._run_subs(
                sub_type=self.SUB_VALUE,
                value=val,
                timestamp=time.time(),
            )
        return val


class GalilAxesReferenced(GalilSignalRO):
    @threadlocked
    def _socket_get(self):
        return self.controller.socket_put_and_receive("MG allaxref")


class SGalilMotor(Device, PositionerBase):
    """ "SGalil Motors at cSAXS have a
    DC motor (y axis - vertical) - implemented as C
    and a step motor (x-axis horizontal) - implemented as E
    that require different communication for control
    """

    USER_ACCESS = ["controller"]
    readback = Cpt(
        GalilReadbackSignal,
        signal_name="readback",
        kind="hinted",
    )
    user_setpoint = Cpt(GalilSetpointSignal, signal_name="setpoint")
    motor_is_moving = Cpt(GalilMotorIsMoving, signal_name="motor_is_moving", kind="normal")
    all_axes_referenced = Cpt(GalilAxesReferenced, signal_name="all_axes_referenced", kind="config")
    high_limit_travel = Cpt(Signal, value=0, kind="omitted")
    low_limit_travel = Cpt(Signal, value=0, kind="omitted")

    SUB_READBACK = "readback"
    SUB_CONNECTION_CHANGE = "connection_change"
    _default_sub = SUB_READBACK

    def __init__(
        self,
        axis_Id,
        prefix="",
        *,
        name,
        kind=None,
        read_attrs=None,
        configuration_attrs=None,
        parent=None,
        host="129.129.122.26",
        port=23,
        limits=None,
        sign=1,
        socket_cls=SocketIO,
        device_manager=None,
        **kwargs,
    ):
        self.axis_Id = axis_Id
        self.sign = sign
        self.controller = GalilController(socket=socket_cls(host=host, port=port))
        self.controller.set_axis(axis=self, axis_nr=self.axis_Id_numeric)
        self.tolerance = kwargs.pop("tolerance", 0.5)
        self.device_mapping = kwargs.pop("device_mapping", {})
        self.device_manager = device_manager

        if len(self.device_mapping) > 0 and self.device_manager is None:
            raise BECConfigError(
                "device_mapping has been specified but the device_manager cannot be accessed."
            )
        self.rt = self.device_mapping.get("rt")

        super().__init__(
            prefix,
            name=name,
            kind=kind,
            read_attrs=read_attrs,
            configuration_attrs=configuration_attrs,
            parent=parent,
            **kwargs,
        )
        self.readback.name = self.name
        self.controller.subscribe(
            self._update_connection_state, event_type=self.SUB_CONNECTION_CHANGE
        )
        self._update_connection_state()
        # self.readback.subscribe(self._forward_readback, event_type=self.readback.SUB_VALUE)

        if limits is not None:
            assert len(limits) == 2
            self.low_limit_travel.put(limits[0])
            self.high_limit_travel.put(limits[1])

    @property
    def limits(self):
        return (self.low_limit_travel.get(), self.high_limit_travel.get())

    @property
    def low_limit(self):
        return self.limits[0]

    @property
    def high_limit(self):
        return self.limits[1]

    def check_value(self, pos):
        """Check that the position is within the soft limits"""
        low_limit, high_limit = self.limits

        if low_limit < high_limit and not (low_limit <= pos <= high_limit):
            raise LimitError(f"position={pos} not within limits {self.limits}")

    def _update_connection_state(self, **kwargs):
        for walk in self.walk_signals():
            walk.item._metadata["connected"] = self.controller.connected

    def _forward_readback(self, **kwargs):
        kwargs.pop("sub_type")
        self._run_subs(sub_type="readback", **kwargs)

    @raise_if_disconnected
    def move(self, position, wait=True, **kwargs):
        """Move to a specified position, optionally waiting for motion to
        complete.

        Parameters
        ----------
        position
            Position to move to
        moved_cb : callable
            Call this callback when movement has finished. This callback must
            accept one keyword argument: 'obj' which will be set to this
            positioner instance.
        timeout : float, optional
            Maximum time to wait for the motion. If None, the default timeout
            for this positioner is used.

        Returns
        -------
        status : MoveStatus

        Raises
        ------
        TimeoutError
            When motion takes longer than `timeout`
        ValueError
            On invalid positions
        RuntimeError
            If motion fails other than timing out
        """
        self._started_moving = False
        timeout = kwargs.pop("timeout", 100)
        status = super().move(position, timeout=timeout, **kwargs)
        self.user_setpoint.put(position, wait=False)

        def move_and_finish():
            while self.motor_is_moving.get():
                logger.info("motor is moving")
                val = self.readback.read()
                self._run_subs(
                    sub_type=self.SUB_READBACK,
                    value=val,
                    timestamp=time.time(),
                )
                time.sleep(1.5)
            val = self.readback.read()
            success = np.isclose(
                val[self.name]["value"],
                position,
                atol=self.tolerance,
            )

            if not success:
                print(" stop")
            self._done_moving(success=success)
            logger.info("Move finished")

        threading.Thread(target=move_and_finish, daemon=True).start()
        try:
            if wait:
                status_wait(status)
        except KeyboardInterrupt:
            self.stop()
            raise

        return status

    @property
    def axis_Id(self):
        return self._axis_Id_alpha

    @axis_Id.setter
    def axis_Id(self, val):
        if isinstance(val, str):
            if len(val) != 1:
                raise ValueError(f"Only single-character axis_Ids are supported.")
            if val not in ["C", "E"]:
                raise ValueError(
                    f"axis_id {val} is currently not supported, please use either 'C' or 'E'."
                )
            self._axis_Id_alpha = val
            self._axis_Id_numeric = ord(val.lower()) - 97
        else:
            raise TypeError(f"Expected value of type str but received {type(val)}")

    @property
    def axis_Id_numeric(self):
        return self._axis_Id_numeric

    @axis_Id_numeric.setter
    def axis_Id_numeric(self, val):
        if isinstance(val, int):
            if val not in [2, 4]:
                raise ValueError(f"Numeric value {val} is not supported, it must be either 2 or 4.")
            self._axis_Id_alpha = val
            self._axis_Id_numeric = (chr(val + 97)).capitalize()
        else:
            raise TypeError(f"Expected value of type int but received {type(val)}.")

    @property
    def egu(self):
        """The engineering units (EGU) for positions"""
        return "mm"

    def stop(self, *, success=False):
        self.controller.stop_all_axes()
        # last_speed = self.controller.socket_put_and_receive("MG")
        rtr = self.controller.socket_put_and_receive(f"SPC={2*10000}")
        logger.info(f"{rtr}")
        # logger.info(f'Motor stopped, restored speed for samy from {last_speed}mm/s to 2mm/s')
        return super().stop(success=success)

    def kickoff(self) -> DeviceStatus:
        status = DeviceStatus(self)
        self.controller.fly_grid_scan(
            status,
            self._kickoff_params.get("start_y"),
            self._kickoff_params.get("end_y"),
            self._kickoff_params.get("interval_y"),
            self._kickoff_params.get("start_x"),
            self._kickoff_params.get("end_x"),
            self._kickoff_params.get("interval_x"),
            self._kickoff_params.get("exp_time"),
            self._kickoff_params.get("readout_time"),
        )
        return status

    def configure(
        self,
        parameter: dict,
        **kwargs,
    ) -> None:
        self._kickoff_params = parameter


if __name__ == "__main__":
    mock = False
    if not mock:
        samy = SGalilMotor("C", name="samy", host="129.129.122.26", port=23, sign=-1)
        samx = SGalilMotor("E", name="samx", host="129.129.122.26", port=23, sign=-1)
    else:
        from ophyd_devices.utils.socket import SocketMock

        samx = SGalilMotor("E", name="samx", host="129.129.122.26", port=23, socket_cls=SocketMock)
        samy = SGalilMotor("C", name="samy", host="129.129.122.26", port=23, socket_cls=SocketMock)

        samx.controller.galil_show_all()



import enum
import threading
import time
from typing import Any, List
import numpy as np

from ophyd import EpicsSignal, EpicsSignalRO
from ophyd import EpicsSignal, EpicsSignalRO, Component as Cpt, Device
from ophyd.mca import EpicsMCARecord
from ophyd.scaler import ScalerCH

from ophyd_devices.epics.devices.bec_scaninfo_mixin import BecScaninfoMixin
from ophyd_devices.utils import bec_utils

from bec_lib import messages, MessageEndpoints, bec_logger, threadlocked
from bec_lib.file_utils import FileWriterMixin
from collections import defaultdict

logger = bec_logger.logger


class McsError(Exception):
    pass


class McsTimeoutError(Exception):
    pass


class TriggerSource(int, enum.Enum):
    MODE0 = 0
    MODE1 = 1
    MODE2 = 2
    MODE3 = 3
    MODE4 = 4
    MODE5 = 5
    MODE6 = 6


class ChannelAdvance(int, enum.Enum):
    INTERNAL = 0
    EXTERNAL = 1


class ReadoutMode(int, enum.Enum):
    PASSIVE = 0
    EVENT = 1
    IO_INTR = 2
    FREQ_0_1HZ = 3
    FREQ_0_2HZ = 4
    FREQ_0_5HZ = 5
    FREQ_1HZ = 6
    FREQ_2HZ = 7
    FREQ_5HZ = 8
    FREQ_10HZ = 9
    FREQ_100HZ = 10


class SIS38XX(Device):
    """SIS38XX control"""

    # Acquisition
    erase_all = Cpt(EpicsSignal, "EraseAll")
    erase_start = Cpt(EpicsSignal, "EraseStart", trigger_value=1)
    start_all = Cpt(EpicsSignal, "StartAll")
    stop_all = Cpt(EpicsSignal, "StopAll")

    acquiring = Cpt(EpicsSignal, "Acquiring")

    preset_real = Cpt(EpicsSignal, "PresetReal")
    elapsed_real = Cpt(EpicsSignal, "ElapsedReal")

    read_mode = Cpt(EpicsSignal, "ReadAll.SCAN")
    read_all = Cpt(EpicsSignal, "DoReadAll.VAL", trigger_value=1)
    num_use_all = Cpt(EpicsSignal, "NuseAll")
    current_channel = Cpt(EpicsSignal, "CurrentChannel")
    dwell = Cpt(EpicsSignal, "Dwell")
    channel_advance = Cpt(EpicsSignal, "ChannelAdvance")
    count_on_start = Cpt(EpicsSignal, "CountOnStart")
    software_channel_advance = Cpt(EpicsSignal, "SoftwareChannelAdvance")
    channel1_source = Cpt(EpicsSignal, "Channel1Source")
    prescale = Cpt(EpicsSignal, "Prescale")
    enable_client_wait = Cpt(EpicsSignal, "EnableClientWait")
    client_wait = Cpt(EpicsSignal, "ClientWait")
    acquire_mode = Cpt(EpicsSignal, "AcquireMode")
    mux_output = Cpt(EpicsSignal, "MUXOutput")
    user_led = Cpt(EpicsSignal, "UserLED")
    input_mode = Cpt(EpicsSignal, "InputMode")
    input_polarity = Cpt(EpicsSignal, "InputPolarity")
    output_mode = Cpt(EpicsSignal, "OutputMode")
    output_polarity = Cpt(EpicsSignal, "OutputPolarity")
    model = Cpt(EpicsSignalRO, "Model", string=True)
    firmware = Cpt(EpicsSignalRO, "Firmware")
    max_channels = Cpt(EpicsSignalRO, "MaxChannels")


class McsCsaxs(SIS38XX):
    USER_ACCESS = ["_init_mcs"]
    SUB_PROGRESS = "progress"
    SUB_VALUE = "value"
    _default_sub = SUB_VALUE
    # scaler = Cpt(ScalerCH, "scaler1")

    # mca2 = Cpt(EpicsMCARecord, "mca2")
    mca1 = Cpt(EpicsSignalRO, "mca1.VAL", auto_monitor=True)
    mca3 = Cpt(EpicsSignalRO, "mca3.VAL", auto_monitor=True)
    mca4 = Cpt(EpicsSignalRO, "mca4.VAL", auto_monitor=True)
    # mca5 = Cpt(EpicsMCARecord, "mca5")
    # mca6 = Cpt(EpicsMCARecord, "mca6")
    # mca7 = Cpt(EpicsMCARecord, "mca7")
    # mca8 = Cpt(EpicsMCARecord, "mca8")
    # mca9 = Cpt(EpicsMCARecord, "mca9")
    # mca10 = Cpt(EpicsMCARecord, "mca10")
    # mca11 = Cpt(EpicsMCARecord, "mca11")
    # mca12 = Cpt(EpicsMCARecord, "mca12")
    # mca13 = Cpt(EpicsMCARecord, "mca13")
    # mca14 = Cpt(EpicsMCARecord, "mca14")
    # mca15 = Cpt(EpicsMCARecord, "mca15")
    # mca16 = Cpt(EpicsMCARecord, "mca16")
    # mca17 = Cpt(EpicsMCARecord, "mca17")
    # mca18 = Cpt(EpicsMCARecord, "mca18")
    # mca19 = Cpt(EpicsMCARecord, "mca19")
    # mca20 = Cpt(EpicsMCARecord, "mca20")
    # mca21 = Cpt(EpicsMCARecord, "mca21")
    # mca22 = Cpt(EpicsMCARecord, "mca22")
    # mca23 = Cpt(EpicsMCARecord, "mca23")
    # mca24 = Cpt(EpicsMCARecord, "mca24")
    # mca25 = Cpt(EpicsMCARecord, "mca25")
    # mca26 = Cpt(EpicsMCARecord, "mca26")
    # mca27 = Cpt(EpicsMCARecord, "mca27")
    # mca28 = Cpt(EpicsMCARecord, "mca28")
    # mca29 = Cpt(EpicsMCARecord, "mca29")
    # mca30 = Cpt(EpicsMCARecord, "mca30")
    # mca31 = Cpt(EpicsMCARecord, "mca31")
    # mca32 = Cpt(EpicsMCARecord, "mca32")
    current_channel = Cpt(EpicsSignalRO, "CurrentChannel", auto_monitor=True)

    num_lines = Cpt(
        bec_utils.ConfigSignal,
        name="num_lines",
        kind="config",
        config_storage_name="mcs_config",
    )

    def __init__(
        self,
        prefix="",
        *,
        name,
        kind=None,
        read_attrs=None,
        configuration_attrs=None,
        parent=None,
        device_manager=None,
        sim_mode=False,
        mcs_config=None,
        **kwargs,
    ):
        self.mcs_config = {
            f"{name}_num_lines": 1,
        }
        if mcs_config is not None:
            [self.mcs_config.update({f"{name}_{key}": value}) for key, value in mcs_config.items()]

        super().__init__(
            prefix=prefix,
            name=name,
            kind=kind,
            read_attrs=read_attrs,
            configuration_attrs=configuration_attrs,
            parent=parent,
            **kwargs,
        )
        if device_manager is None and not sim_mode:
            raise McsError("Add DeviceManager to initialization or init with sim_mode=True")

        self.name = name
        self._stream_ttl = 1800
        self.wait_for_connection()  # Make sure to be connected before talking to PVs

        if not sim_mode:
            self.device_manager = device_manager
            self._producer = self.device_manager.producer
        else:
            self._producer = bec_utils.MockProducer()
            self.device_manager = bec_utils.MockDeviceManager()
            # TODO mack mock connector class
            # self._consumer = self.device_manager.connector.consumer
        self.scaninfo = BecScaninfoMixin(device_manager, sim_mode)
        # TODO
        self.scaninfo.username = "e21206"
        self.service_cfg = {"base_path": f"/sls/X12SA/data/{self.scaninfo.username}/Data10/"}
        self.filewriter = FileWriterMixin(self.service_cfg)
        self._stopped = False
        self._acquisition_done = False
        self._lock = threading.RLock()
        self.counter = 0
        self.n_points = 0
        self._init_mcs()
        self.mca_names = [signal for signal in self.component_names if signal.startswith("mca")]
        self.mca_data = defaultdict(lambda: [])
        for mca in self.mca_names:
            signal = getattr(self, mca)
            signal.subscribe(self._on_mca_data, run=False)
        self.current_channel.subscribe(self._progress_update, run=False)

    def _init_mcs(self) -> None:
        """Init parameters for mcs card 9m
        channel_advance: 0/1 -> internal / external
        channel1_source: 0/1 -> int clock / external source
        user_led: 0/1 -> off/on
        max_output :  num of channels 0...32, uncomment top for more than 5
        input_mode: operation mode -> Mode 3 for external trigger, check manual for more info
        input_polarity: triggered between falling and falling edge -> use inverted signal from ddg
        """
        self.channel_advance.set(ChannelAdvance.EXTERNAL)
        self.channel1_source.set(ChannelAdvance.INTERNAL)
        self.user_led.set(0)
        self.mux_output.set(5)
        self._set_trigger(TriggerSource.MODE3)
        self.input_polarity.set(0)
        self.output_polarity.set(1)
        self.count_on_start.set(0)
        self.stop_all.set(1)

    def _progress_update(self, value, **kwargs) -> None:
        num_lines = self.num_lines.get()
        max_value = self.scaninfo.num_points
        self._run_subs(
            sub_type=self.SUB_PROGRESS,
            value=self.counter * int(self.scaninfo.num_points / num_lines) + value,
            max_value=max_value,
            done=bool(max_value == self.counter),
        )

    @threadlocked
    def _on_mca_data(self, *args, obj=None, **kwargs) -> None:
        if not isinstance(kwargs["value"], (list, np.ndarray)):
            return
        # self.mca_data[obj.attr_name] = kwargs["value"][1:]
        self.mca_data[obj.attr_name] = kwargs["value"]
        if len(self.mca_names) != len(self.mca_data):
            return
        # logger.info("Entered _on_mca_data")
        # self._updated = True
        # self.counter += 1
        # logger.info(f'data from mca {self.mca_data["mca1"]} and {self.mca_data["mca4"]}')
        # if (self.scaninfo.scan_type == "fly" and self.counter == self.num_lines.get()) or (
        #     self.scaninfo.scan_type == "step" and self.counter == self.scaninfo.num_points
        # ):
        #     self._acquisition_done = True
        #     self.stop_all.put(1, use_complete=False)
        #     self._send_data_to_bec()
        #     self.erase_all.put(1)
        #     #logger.info("Entered _on_mca_data, acquisition finished")
        #     # Require wait for
        #     # time.sleep(0.01)
        #     self.mca_data = defaultdict(lambda: [])
        #     self.counter = 0
        #     return
        # self.erase_start.set(1)
        # self._send_data_to_bec()
        # self.mca_data = defaultdict(lambda: [])
        self._acquisition_done = True
        self._send_data_to_bec()
        self.mca_data = defaultdict(lambda: [])

    def _send_data_to_bec(self) -> None:
        if self.scaninfo.scan_msg is None:
            return
        metadata = self.scaninfo.scan_msg.metadata
        metadata.update(
            {
                "async_update": "append",
                "num_lines": self.num_lines.get(),
            }
        )
        msg = messages.DeviceMessage(
            signals=dict(self.mca_data),
            metadata=self.scaninfo.scan_msg.metadata,
        ).dumps()
        self._producer.xadd(
            topic=MessageEndpoints.device_async_readback(
                scanID=self.scaninfo.scanID, device=self.name
            ),
            msg={"data": msg},
            expire=self._stream_ttl,
        )

    def _prep_det(self) -> None:
        self._set_acquisition_params()
        self._set_trigger(TriggerSource.MODE3)

    def _set_acquisition_params(self) -> None:
        if self.scaninfo.scan_type == "step":
            self.n_points = int(self.scaninfo.frames_per_trigger) * int(self.scaninfo.num_points)
        elif self.scaninfo.scan_type == "fly":
            self.n_points = int(self.scaninfo.num_points)  # / int(self.num_lines.get()))
        else:
            raise McsError(f"Scantype {self.scaninfo} not implemented for MCS card")
        if self.n_points > 10000:
            raise McsError(
                f"Requested number of points N={self.n_points} exceeds hardware limit of mcs card 10000 (N-1)"
            )
        self.num_use_all.set(self.n_points)
        self.preset_real.set(0)

    def _set_trigger(self, trigger_source: TriggerSource) -> None:
        """7 Modes, see TriggerSource
        Mode3 for cSAXS"""
        value = int(trigger_source)
        self.input_mode.set(value)

    def _prep_readout(self) -> None:
        """Set readout mode of mcs card
        Check ReadoutMode class for more information about options
        """
        # self.read_mode.set(ReadoutMode.EVENT)
        self.erase_all.set(1)
        self.read_mode.set(ReadoutMode.EVENT)

    def _force_readout_mcs_card(self) -> None:
        self.read_all.put(1, use_complete=False)

    def stage(self) -> List[object]:
        """stage the detector and file writer"""
        self._stopped = False
        self._acquisition_done = False
        logger.info("Stage mcs")
        self.scaninfo.load_scan_metadata()
        self._prep_det()
        self._prep_readout()

        # msg = messages.FileMessage(file_path=self.filepath, done=False)
        # self._producer.set_and_publish(
        #     MessageEndpoints.public_file(self.scaninfo.scanID, "mcs_csaxs"),
        #     msg.dumps(),
        # )
        self.arm_acquisition()
        logger.info("Waiting for mcs to be armed")
        while True:
            det_ctrl = self.acquiring.read()[self.acquiring.name]["value"]
            if det_ctrl == 1:
                break
            time.sleep(0.005)
        logger.info("mcs is ready and running")
        # time.sleep(5)
        return super().stage()

    def unstage(self) -> List[object]:
        """unstage"""
        logger.info("Waiting for mcs to finish acquisition")
        old_scanID = self.scaninfo.scanID
        self.scaninfo.load_scan_metadata()
        logger.info(f"Old scanID: {old_scanID}, ")
        if self.scaninfo.scanID != old_scanID:
            self._stopped = True
        if self._stopped is True:
            logger.info("Entered unstage _stopped =True")
            return super().unstage()
        self._mcs_finished()
        self._acquisition_done = False
        self._stopped = False
        logger.info("mcs done")
        return super().unstage()

    def _mcs_finished(self):
        """Function with 10s timeout"""
        timer = 0
        logger.info("Entered _mcs_finished loop")
        while True:
            if self._acquisition_done == True and self.acquiring.get() == 0:
                break
            if self._stopped == True:
                break
            time.sleep(0.1)
            timer += 0.1
            if timer > 8:
                total_frames = self.counter * int(
                    self.scaninfo.num_points / self.num_lines.get()
                ) + max(self.current_channel.get(), 0)
                raise McsTimeoutError(
                    f"Reached timeout with mcs in state {self.acquiring.get()} and {total_frames} frames arriving at the mcs card"
                )
        logger.info("Finished _mcs_finished loop")

    def arm_acquisition(self) -> None:
        """Arm acquisition
        Options:
        Start: start_all
        Erase/Start: erase_start
        """
        logger.info("Entered mcs arm_acquisition")
        self.counter = 0
        self.erase_start.set(1)
        # self.start_all.set(1)

    def stop(self, *, success=False) -> None:
        """Stop acquisition
        Stop or Stop and Erase
        """
        logger.info("Entered mcs stop")
        self.stop_all.set(1)
        # self.erase_all.set(1)
        self._stopped = True
        self._acquisition_done = True
        super().stop(success=success)


# Automatically connect to test environmenr if directly invoked
if __name__ == "__main__":
    mcs = McsCsaxs(name="mcs", prefix="X12SA-MCS:", sim_mode=True)
    mcs.stage()
    mcs.unstage()



from setuptools import setup

__version__ = "0.13.2"

if __name__ == "__main__":
    setup(
        install_requires=[
            "ophyd",
            "typeguard",
            "prettytable",
            "bec_lib",
            "numpy",
            "pyyaml",
            "std_daq_client",
            "pyepics",
            "pytest",
        ],
        extras_require={"dev": ["pytest", "pytest-random-order", "black", "coverage"]},
        package_data={"ophyd_devices.smaract": ["smaract_sensors.json"]},
        version=__version__,
    )



# Standard ophyd classes
from ophyd import EpicsMotor, EpicsSignal, EpicsSignalRO
from ophyd.quadem import QuadEM
from ophyd.sim import SynAxis, SynPeriodicSignal, SynSignal

from .devices.DelayGeneratorDG645 import DelayGeneratorDG645
from .devices.InsertionDevice import InsertionDevice
from .devices.slits import SlitH, SlitV
from .devices.specMotors import (
    Bpm4i,
    EnergyKev,
    GirderMotorPITCH,
    GirderMotorROLL,
    GirderMotorX1,
    GirderMotorY1,
    GirderMotorYAW,
    MonoTheta1,
    MonoTheta2,
    PmDetectorRotation,
    PmMonoBender,
)
from .devices.SpmBase import SpmBase
from .devices.XbpmBase import XbpmBase, XbpmCsaxsOp

# X07MA specific devices
from .devices.X07MADevices import *



from .fgalil_ophyd import FlomniGalilController, FlomniGalilMotor
from .fupr_ophyd import FuprGalilController, FuprGalilMotor
from .galil_ophyd import GalilController, GalilMotor
from .sgalil_ophyd import SGalilMotor



import enum
import threading
import time
from typing import Any, List
from ophyd import Device, Component, EpicsSignal, EpicsSignalRO, Kind
from ophyd import PVPositioner, Signal, DeviceStatus
from ophyd.pseudopos import (
    pseudo_position_argument,
    real_position_argument,
    PseudoSingle,
    PseudoPositioner,
)
from ophyd_devices.utils.socket import data_shape, data_type
from ophyd_devices.utils import bec_utils as bec_utils

from bec_lib import bec_logger

from ophyd_devices.epics.devices.bec_scaninfo_mixin import BecScaninfoMixin


logger = bec_logger.logger


class DDGError(Exception):
    pass


class DelayStatic(Device):
    """Static axis for the T0 output channel
    It allows setting the logic levels, but the timing is fixed.
    The signal is high after receiving the trigger until the end
    of the holdoff period.
    """

    # Other channel stuff
    ttl_mode = Component(EpicsSignal, "OutputModeTtlSS.PROC", kind=Kind.config)
    nim_mode = Component(EpicsSignal, "OutputModeNimSS.PROC", kind=Kind.config)
    polarity = Component(
        EpicsSignal,
        "OutputPolarityBI",
        write_pv="OutputPolarityBO",
        name="polarity",
        kind=Kind.config,
    )
    amplitude = Component(
        EpicsSignal,
        "OutputAmpAI",
        write_pv="OutputAmpAO",
        name="amplitude",
        kind=Kind.config,
    )
    offset = Component(
        EpicsSignal,
        "OutputOffsetAI",
        write_pv="OutputOffsetAO",
        name="offset",
        kind=Kind.config,
    )


class DummyPositioner(PVPositioner):
    setpoint = Component(EpicsSignal, "DelayAO", put_complete=True, kind=Kind.config)
    readback = Component(EpicsSignalRO, "DelayAI", kind=Kind.config)
    done = Component(Signal, value=1)
    reference = Component(EpicsSignal, "ReferenceMO", put_complete=True, kind=Kind.config)


class DelayPair(PseudoPositioner):
    """Delay pair interface for DG645

    Virtual motor interface to a pair of signals (on the frontpanel).
    It offers a simple delay and pulse width interface for scanning.
    """

    # The pseudo positioner axes
    delay = Component(PseudoSingle, limits=(0, 2000.0), name="delay")
    width = Component(PseudoSingle, limits=(0, 2000.0), name="pulsewidth")
    ch1 = Component(DummyPositioner, name="ch1")
    ch2 = Component(DummyPositioner, name="ch2")
    io = Component(DelayStatic, name="io")

    def __init__(self, *args, **kwargs):
        # Change suffix names before connecting (a bit of dynamic connections)
        self.__class__.__dict__["ch1"].suffix = kwargs["channel"][0]
        self.__class__.__dict__["ch2"].suffix = kwargs["channel"][1]
        self.__class__.__dict__["io"].suffix = kwargs["channel"]

        del kwargs["channel"]
        # Call parent to start the connections
        super().__init__(*args, **kwargs)

    @pseudo_position_argument
    def forward(self, pseudo_pos):
        """Run a forward (pseudo -> real) calculation"""
        return self.RealPosition(ch1=pseudo_pos.delay, ch2=pseudo_pos.delay + pseudo_pos.width)

    @real_position_argument
    def inverse(self, real_pos):
        """Run an inverse (real -> pseudo) calculation"""
        return self.PseudoPosition(delay=real_pos.ch1, width=real_pos.ch2 - real_pos.ch1)


class TriggerSource(int, enum.Enum):
    INTERNAL = 0
    EXT_RISING_EDGE = 1
    EXT_FALLING_EDGE = 2
    SS_EXT_RISING_EDGE = 3
    SS_EXT_FALLING_EDGE = 4
    SINGLE_SHOT = 5
    LINE = 6


class DelayGeneratorDG645(Device):
    """DG645 delay generator

    This class implements a thin Ophyd wrapper around the Stanford Research DG645
    digital delay generator.

    Internally, the DG645 generates 8+1 signals:  A, B, C, D, E, F, G, H and T0
    Front panel outputs T0, AB, CD, EF and GH are a combination of these signals.
    Back panel outputs are directly routed signals. So signals are NOT INDEPENDENT.

    Front panel signals:
    All signals go high after their defined delays and go low after the trigger
    holdoff period, i.e. this is the trigger window. Front panel outputs provide
    a combination of these events.
    Option 1 back panel 5V signals:
    All signals go high after their defined delays and go low after the trigger
    holdoff period, i.e. this is the trigger window. The signals will stay high
    until the end of the window.
    Option 2 back panel 30V signals:
    All signals go high after their defined delays for ~100ns. This is fixed by
    electronics (30V needs quite some power). This is not implemented in the
    current device
    """

    SUB_PROGRESS = "progress"
    SUB_VALUE = "value"
    _default_sub = SUB_VALUE

    USER_ACCESS = [
        "set_channels",
        "_set_trigger",
        "burst_enable",
        "burst_disable",
        "reload_config",
    ]

    trigger_burst_readout = Component(
        EpicsSignal, "EventStatusLI.PROC", name="trigger_burst_readout"
    )
    burst_cycle_finished = Component(EpicsSignalRO, "EventStatusMBBID.B3", name="read_burst_state")
    delay_finished = Component(EpicsSignalRO, "EventStatusMBBID.B2", name="delay_finished")
    status = Component(EpicsSignalRO, "StatusSI", name="status")
    clear_error = Component(EpicsSignal, "StatusClearBO", name="clear_error")

    # Front Panel
    channelT0 = Component(DelayStatic, "T0", name="T0")
    channelAB = Component(DelayPair, "", name="AB", channel="AB")
    channelCD = Component(DelayPair, "", name="CD", channel="CD")
    channelEF = Component(DelayPair, "", name="EF", channel="EF")
    channelGH = Component(DelayPair, "", name="GH", channel="GH")

    # Minimum time between triggers
    holdoff = Component(
        EpicsSignal,
        "TriggerHoldoffAI",
        write_pv="TriggerHoldoffAO",
        name="trigger_holdoff",
        kind=Kind.config,
    )
    inhibit = Component(
        EpicsSignal,
        "TriggerInhibitMI",
        write_pv="TriggerInhibitMO",
        name="trigger_inhibit",
        kind=Kind.config,
    )
    source = Component(
        EpicsSignal,
        "TriggerSourceMI",
        write_pv="TriggerSourceMO",
        name="trigger_source",
        kind=Kind.config,
    )
    level = Component(
        EpicsSignal,
        "TriggerLevelAI",
        write_pv="TriggerLevelAO",
        name="trigger_level",
        kind=Kind.config,
    )
    rate = Component(
        EpicsSignal,
        "TriggerRateAI",
        write_pv="TriggerRateAO",
        name="trigger_rate",
        kind=Kind.config,
    )
    trigger_shot = Component(EpicsSignal, "TriggerDelayBO", name="trigger_shot", kind="config")
    # Burst mode
    burstMode = Component(
        EpicsSignal,
        "BurstModeBI",
        write_pv="BurstModeBO",
        name="burstmode",
        kind=Kind.config,
    )
    burstConfig = Component(
        EpicsSignal,
        "BurstConfigBI",
        write_pv="BurstConfigBO",
        name="burstconfig",
        kind=Kind.config,
    )
    burstCount = Component(
        EpicsSignal,
        "BurstCountLI",
        write_pv="BurstCountLO",
        name="burstcount",
        kind=Kind.config,
    )
    burstDelay = Component(
        EpicsSignal,
        "BurstDelayAI",
        write_pv="BurstDelayAO",
        name="burstdelay",
        kind=Kind.config,
    )
    burstPeriod = Component(
        EpicsSignal,
        "BurstPeriodAI",
        write_pv="BurstPeriodAO",
        name="burstperiod",
        kind=Kind.config,
    )

    delay_burst = Component(
        bec_utils.ConfigSignal,
        name="delay_burst",
        kind="config",
        config_storage_name="ddg_config",
    )

    delta_width = Component(
        bec_utils.ConfigSignal,
        name="delta_width",
        kind="config",
        config_storage_name="ddg_config",
    )

    additional_triggers = Component(
        bec_utils.ConfigSignal,
        name="additional_triggers",
        kind="config",
        config_storage_name="ddg_config",
    )

    polarity = Component(
        bec_utils.ConfigSignal,
        name="polarity",
        kind="config",
        config_storage_name="ddg_config",
    )

    fixed_ttl_width = Component(
        bec_utils.ConfigSignal,
        name="fixed_ttl_width",
        kind="config",
        config_storage_name="ddg_config",
    )

    amplitude = Component(
        bec_utils.ConfigSignal,
        name="amplitude",
        kind="config",
        config_storage_name="ddg_config",
    )

    offset = Component(
        bec_utils.ConfigSignal,
        name="offset",
        kind="config",
        config_storage_name="ddg_config",
    )

    thres_trig_level = Component(
        bec_utils.ConfigSignal,
        name="thres_trig_level",
        kind="config",
        config_storage_name="ddg_config",
    )

    set_high_on_exposure = Component(
        bec_utils.ConfigSignal,
        name="set_high_on_exposure",
        kind="config",
        config_storage_name="ddg_config",
    )

    set_high_on_stage = Component(
        bec_utils.ConfigSignal,
        name="set_high_on_stage",
        kind="config",
        config_storage_name="ddg_config",
    )

    set_trigger_source = Component(
        bec_utils.ConfigSignal,
        name="set_trigger_source",
        kind="config",
        config_storage_name="ddg_config",
    )

    trigger_width = Component(
        bec_utils.ConfigSignal,
        name="trigger_width",
        kind="config",
        config_storage_name="ddg_config",
    )
    premove_trigger = Component(
        bec_utils.ConfigSignal,
        name="premove_trigger",
        kind="config",
        config_storage_name="ddg_config",
    )

    def __init__(
        self,
        prefix="",
        *,
        name,
        kind=None,
        read_attrs=None,
        configuration_attrs=None,
        parent=None,
        device_manager=None,
        sim_mode=False,
        ddg_config=None,
        **kwargs,
    ):
        """_summary_

        Args:
            name (_type_): _description_
            prefix (str, optional): _description_. Defaults to "".
            kind (_type_, optional): _description_. Defaults to None.
            read_attrs (_type_, optional): _description_. Defaults to None.
            configuration_attrs (_type_, optional): _description_. Defaults to None.
            parent (_type_, optional): _description_. Defaults to None.
            device_manager (_type_, optional): _description_. Defaults to None.
        Signals:
            polarity (_list_, optional): _description_. Defaults to None.
            fixed_ttl_width (_list_, optional): _description_. Defaults to None.
            amplitude (_type_, optional): _description_. Defaults to None.
            offset (_type_, optional): _description_. Defaults to None.
            thres_trig_level (_type_, optional): _description_. Defaults to None.
            delay_burst (_type_, float): Add delay for triggering in software trigger mode to allow fast shutter to open. Defaults to 0.
            delta_width (_type_, float): Add width to fast shutter signal to make sure its open during acquisition. Defaults to 0.
            delta_triggers (_type_, int): Add additional triggers to burst mode (mcs card needs +1 triggers per line). Defaults to 0.
            set_high_on_exposure
            set_high_on_stage
            set_trigger_source

        """
        self.ddg_config = {
            f"{name}_delay_burst": 0,
            f"{name}_delta_width": 0,
            f"{name}_additional_triggers": 0,
            f"{name}_polarity": [1, 1, 1, 1, 1],
            f"{name}_fixed_ttl_width": [0, 0, 0, 0, 0],
            f"{name}_amplitude": 4.5,
            f"{name}_offset": 0,
            f"{name}_thres_trig_level": 2.5,
            f"{name}_set_high_on_exposure": False,
            f"{name}_set_high_on_stage": False,
            f"{name}_set_trigger_source": "SINGLE_SHOT",
            f"{name}_trigger_width": None,  # This somehow duplicates the logic of fixed_ttl_width
            f"{name}_premove_trigger": False,
        }
        if ddg_config is not None:
            [self.ddg_config.update({f"{name}_{key}": value}) for key, value in ddg_config.items()]
        super().__init__(
            prefix=prefix,
            name=name,
            kind=kind,
            read_attrs=read_attrs,
            configuration_attrs=configuration_attrs,
            parent=parent,
            **kwargs,
        )
        if device_manager is None and not sim_mode:
            raise DDGError("Add DeviceManager to initialization or init with sim_mode=True")
        self.device_manager = device_manager
        if not sim_mode:
            self._producer = self.device_manager.producer
        else:
            self._producer = bec_utils.MockProducer()
            self.device_manager = bec_utils.MockDeviceManager()
        self.scaninfo = BecScaninfoMixin(device_manager, sim_mode)
        self._all_channels = [
            "channelT0",
            "channelAB",
            "channelCD",
            "channelEF",
            "channelGH",
        ]
        self._all_delay_pairs = ["AB", "CD", "EF", "GH"]
        self.wait_for_connection()  # Make sure to be connected before talking to PVs
        logger.info(f"Current polarity values {self.polarity.get()}")
        self.reload_config()
        self._ddg_is_okay()
        self._stopped = False

    def _set_trigger(self, trigger_source: TriggerSource) -> None:
        """Set trigger source to value of list below, or string
        Accepts integer 0-6 or TriggerSource.* with *
        INTERNAL = 0
        EXT_RISING_EDGE = 1
        EXT_FALLING_EDGE = 2
        SS_EXT_RISING_EDGE = 3
        SS_EXT_FALLING_EDGE = 4
        SINGLE_SHOT = 5
        LINE = 6
        """
        value = int(trigger_source)
        self.source.put(value)

    def _ddg_is_okay(self, raise_on_error=False) -> None:
        status = self.status.read()[self.status.name]["value"]
        if status != "STATUS OK" and not raise_on_error:
            logger.warning(f"DDG returns {status}, trying to clear ERROR")
            self.clear_error()
            time.sleep(1)
            self._ddg_is_okay(rais_on_error=True)
        elif status != "STATUS OK":
            raise DDGError(f"DDG failed to start with status: {status}")

    def set_channels(self, signal: str, value: Any, channels: List = None) -> None:
        if not channels:
            channels = self._all_channels
        for chname in channels:
            channel = getattr(self, chname, None)
            if not channel:
                continue
            if signal in channel.component_names:
                getattr(channel, signal).set(value)
                continue
            if "io" in channel.component_names and signal in channel.io.component_names:
                getattr(channel.io, signal).set(value)

    def _cleanup_ddg(self) -> None:
        self._set_trigger(getattr(TriggerSource, self.set_trigger_source.get()))

    def reload_config(self) -> None:
        for ii, channel in enumerate(self._all_channels):
            self.set_channels("polarity", self.polarity.get()[ii], channels=[channel])
        # Set polarity for eiger inverted!
        # self.set_channels("polarity", 0, channels=["channelAB"])
        self.set_channels("amplitude", self.amplitude.get())
        self.set_channels("offset", self.offset.get())
        # Setup reference
        self.set_channels(
            "reference",
            0,
            [f"channel{self._all_delay_pairs[ii]}.ch1" for ii in range(len(self._all_delay_pairs))],
        )
        for ii in range(len(self._all_delay_pairs)):
            self.set_channels(
                "reference",
                0,
                [f"channel{self._all_delay_pairs[ii]}.ch2"],
            )
        self._set_trigger(getattr(TriggerSource, self.set_trigger_source.get()))
        # Set threshold level for ext. pulses
        self.level.put(self.thres_trig_level.get())

    def _check_burst_cycle(self, status) -> None:
        """Checks burst cycle of delay generator
        Force readout, return value from end of burst cycle
        """
        return status.set_finished()
        while True:
            self.trigger_burst_readout.put(1, use_complete=True)
            if self.burst_cycle_finished.read()[self.burst_cycle_finished.name]["value"] == 1:
                self._acquisition_done = True
                status.set_finished()
                return
            if self._stopped == True:
                status.set_finished()
                break

            time.sleep(0.01)

    def stop(self, success=False):
        """Stops the DDG"""
        self._stopped = True
        self._acquisition_done = True
        super().stop(success=success)

    def stage(self):
        """Trigger the generator by arming to accept triggers"""
        self.scaninfo.load_scan_metadata()
        if self.scaninfo.scan_type == "step":
            # define parameters
            if self.set_high_on_exposure.get():
                self._set_trigger(getattr(TriggerSource, self.set_trigger_source.get()))
                num_burst_cycle = 1 + self.additional_triggers.get()

                exp_time = self.delta_width.get() + self.scaninfo.frames_per_trigger * (
                    self.scaninfo.exp_time + self.scaninfo.readout_time
                )
                total_exposure = exp_time
                delay_burst = self.delay_burst.get()
                self.burst_enable(num_burst_cycle, delay_burst, total_exposure, config="first")
                self.set_channels("delay", 0)
                # Set burst length to half of the experimental time!
                if not self.trigger_width.get():
                    self.set_channels("width", exp_time)
                else:
                    self.set_channels("width", self.trigger_width.get())
                for value, channel in zip(self.fixed_ttl_width.get(), self._all_channels):
                    logger.info(f"{value}")
                    if value != 0:
                        logger.info(f"Setting {value}")
                        self.set_channels("width", value, channels=[channel])
            else:
                self._set_trigger(getattr(TriggerSource, self.set_trigger_source.get()))
                exp_time = self.delta_width.get() + self.scaninfo.exp_time
                total_exposure = exp_time + self.scaninfo.readout_time
                delay_burst = self.delay_burst.get()
                num_burst_cycle = self.scaninfo.frames_per_trigger + self.additional_triggers.get()
                # set parameters in DDG
                self.burst_enable(num_burst_cycle, delay_burst, total_exposure, config="first")
                self.set_channels("delay", 0)
                # Set burst length to half of the experimental time!
                if not self.trigger_width.get():
                    self.set_channels("width", exp_time)
                else:
                    self.set_channels("width", self.trigger_width.get())
        elif self.scaninfo.scan_type == "fly":
            if self.set_high_on_exposure.get():
                # define parameters
                self._set_trigger(getattr(TriggerSource, self.set_trigger_source.get()))
                exp_time = (
                    self.delta_width.get()
                    + self.scaninfo.exp_time * self.scaninfo.num_points
                    + self.scaninfo.readout_time * (self.scaninfo.num_points - 1)
                )
                total_exposure = exp_time
                delay_burst = self.delay_burst.get()
                # self.additional_triggers should be 0 for self.set_high_on_exposure or remove here fully..
                num_burst_cycle = 1 + self.additional_triggers.get()
                # set parameters in DDG
                self.burst_enable(num_burst_cycle, delay_burst, total_exposure, config="first")
                self.set_channels("delay", 0.0)
                # Set burst length to half of the experimental time!
                if not self.trigger_width.get():
                    self.set_channels("width", exp_time)
                else:
                    self.set_channels("width", self.trigger_width.get())
                for value, channel in zip(self.fixed_ttl_width.get(), self._all_channels):
                    logger.info(f"{value}")
                    if value != 0:
                        logger.info(f"Setting {value}")
                        self.set_channels("width", value, channels=[channel])
            else:
                # define parameters
                self._set_trigger(getattr(TriggerSource, self.set_trigger_source.get()))
                exp_time = self.delta_width.get() + self.scaninfo.exp_time
                total_exposure = exp_time + self.scaninfo.readout_time
                delay_burst = self.delay_burst.get()
                num_burst_cycle = self.scaninfo.num_points + self.additional_triggers.get()
                # set parameters in DDG
                self.burst_enable(num_burst_cycle, delay_burst, total_exposure, config="first")
                self.set_channels("delay", 0.0)
                # Set burst length to half of the experimental time!
                if not self.trigger_width.get():
                    self.set_channels("width", exp_time)
                else:
                    self.set_channels("width", self.trigger_width.get())

        else:
            raise DDGError(f"Unknown scan type {self.scaninfo.scan_type}")

        # Check status
        self._ddg_is_okay()
        logger.info("DDG staged")
        super().stage()

    def unstage(self):
        """Stop the trigger generator from accepting triggers"""
        # self._set_trigger(getattr(TriggerSource, self.set_trigger_source.get()))
        # Check status
        self._ddg_is_okay()
        self._stopped = False
        self._acquisition_done = False
        logger.info("DDG unstaged")
        super().unstage()

    def pre_scan(self) -> None:
        if self.premove_trigger.get() == True:
            self.trigger_shot.put(1)

    def trigger(self) -> DeviceStatus:
        status = DeviceStatus(self)
        if self.premove_trigger.get() == True:
            status.set_finished()
            return status
        if self.source.read()[self.source.name]["value"] == int(TriggerSource.SINGLE_SHOT):
            self.trigger_shot.put(1)
        burst_state = threading.Thread(target=self._check_burst_cycle, args=(status,), daemon=True)
        burst_state.start()
        return status

    def burst_enable(self, count, delay, period, config="all"):
        """Enable the burst mode"""
        # Validate inputs
        count = int(count)
        assert count > 0, "Number of bursts must be positive"
        assert delay >= 0, "Burst delay must be larger than 0"
        assert period > 0, "Burst period must be positive"
        assert config in [
            "all",
            "first",
        ], "Supported bust configs are 'all' and 'first'"

        self.burstMode.put(1)
        self.burstCount.put(count)
        self.burstDelay.put(delay)
        self.burstPeriod.put(period)

        if config == "all":
            self.burstConfig.put(0)
        elif config == "first":
            self.burstConfig.put(1)

    def burst_disable(self):
        """Disable the burst mode"""
        self.burstMode.put(0)


# Automatically connect to test environmenr if directly invoked
if __name__ == "__main__":
    dgen = DelayGeneratorDG645("delaygen:DG1:", name="dgen", sim_mode=True)

    # start = time.time()
    # dgen.stage()
    # dgen.trigger()
    # print(f"Time passed for stage and trigger {time.time()-start}s")



import functools
import threading
import time
from typing import List

import numpy as np
from bec_lib import bec_logger
from ophyd import Component as Cpt
from ophyd import Device, PositionerBase, Signal
from ophyd.status import wait as status_wait
from ophyd.utils import LimitError, ReadOnlyError
from prettytable import PrettyTable

from ophyd_devices.utils.controller import Controller, threadlocked
from ophyd_devices.utils.socket import SocketIO, SocketSignal, raise_if_disconnected

logger = bec_logger.logger


class GalilCommunicationError(Exception):
    pass


class GalilError(Exception):
    pass


class BECConfigError(Exception):
    pass


def retry_once(fcn):
    """Decorator to rerun a function in case a Galil communication error was raised. This may happen if the buffer was not empty."""

    @functools.wraps(fcn)
    def wrapper(self, *args, **kwargs):
        try:
            val = fcn(self, *args, **kwargs)
        except (GalilCommunicationError, GalilError):
            val = fcn(self, *args, **kwargs)
        return val

    return wrapper


class GalilController(Controller):
    _axes_per_controller = 8
    USER_ACCESS = [
        "describe",
        "show_running_threads",
        "galil_show_all",
        "socket_put_and_receive",
        "socket_put_confirmed",
        "lgalil_is_air_off_and_orchestra_enabled",
    ]

    @threadlocked
    def socket_put(self, val: str) -> None:
        self.sock.put(f"{val}\r".encode())

    @threadlocked
    def socket_get(self) -> str:
        return self.sock.receive().decode()

    @retry_once
    @threadlocked
    def socket_put_and_receive(self, val: str, remove_trailing_chars=True) -> str:
        self.socket_put(val)
        if remove_trailing_chars:
            return self._remove_trailing_characters(self.sock.receive().decode())
        return self.socket_get()

    @retry_once
    def socket_put_confirmed(self, val: str) -> None:
        """Send message to controller and ensure that it is received by checking that the socket receives a colon.

        Args:
            val (str): Message that should be sent to the socket

        Raises:
            GalilCommunicationError: Raised if the return value is not a colon.

        """
        return_val = self.socket_put_and_receive(val)
        if return_val != ":":
            raise GalilCommunicationError(
                f"Expected return value of ':' but instead received {return_val}"
            )

    def is_axis_moving(self, axis_Id, axis_Id_numeric) -> bool:
        if axis_Id is None and axis_Id_numeric is not None:
            axis_Id = self.axis_Id_numeric_to_alpha(axis_Id_numeric)
        is_moving = bool(float(self.socket_put_and_receive(f"MG_BG{axis_Id}")) != 0)
        backlash_is_active = bool(float(self.socket_put_and_receive(f"MGbcklact[axis]")) != 0)
        return bool(
            is_moving or backlash_is_active or self.is_thread_active(0) or self.is_thread_active(2)
        )

    def is_thread_active(self, thread_id: int) -> bool:
        val = float(self.socket_put_and_receive(f"MG_XQ{thread_id}"))
        if val == -1:
            return False
        return True

    def _remove_trailing_characters(self, var) -> str:
        if len(var) > 1:
            return var.split("\r\n")[0]
        return var

    def stop_all_axes(self) -> str:
        return self.socket_put_and_receive(f"XQ#STOP,1")

    def lgalil_is_air_off_and_orchestra_enabled(self) -> bool:
        # TODO: move this to the LamNI-specific controller
        rt_not_blocked_by_galil = bool(self.socket_put_and_receive(f"MG@OUT[9]"))
        air_off = bool(self.socket_put_and_receive(f"MG@OUT[13]"))
        return rt_not_blocked_by_galil and air_off

    def axis_is_referenced(self, axis_Id_numeric) -> bool:
        return bool(float(self.socket_put_and_receive(f"MG axisref[{axis_Id_numeric}]").strip()))

    def all_axes_referenced(self) -> bool:
        """
        Check if all axes are referenced.
        """
        return bool(float(self.socket_put_and_receive("MG allaxref").strip()))

    def drive_axis_to_limit(self, axis_Id_numeric: int, direction: str) -> None:
        """
        Drive an axis to the limit in a specified direction.

        Args:
            axis_Id_numeric (int): Axis number
            direction (str): Direction in which the axis should be driven to the limit. Either 'forward' or 'reverse'.
        """
        if direction == "forward":
            direction_flag = 1
        elif direction == "reverse":
            direction_flag = -1
        else:
            raise ValueError(f"Invalid direction {direction}")

        self.socket_put_confirmed(f"naxis={axis_Id_numeric}")
        self.socket_put_confirmed(f"ndir={direction_flag}")
        self.socket_put_confirmed("XQ#NEWPAR")
        self.socket_put_confirmed("XQ#FES")
        time.sleep(0.1)
        while self.is_axis_moving(None, axis_Id_numeric):
            time.sleep(0.1)

        axis_Id = self.axis_Id_numeric_to_alpha(axis_Id_numeric)
        # check if we actually hit the limit
        if direction == "forward":
            limit = self.get_motor_limit_switch(axis_Id)[1]
        elif direction == "reverse":
            limit = self.get_motor_limit_switch(axis_Id)[0]

        if not limit:
            raise GalilError(f"Failed to drive axis {axis_Id}/{axis_Id_numeric} to limit.")

    def find_reference(self, axis_Id_numeric: int) -> None:
        """
        Find the reference of an axis.

        Args:
            axis_Id_numeric (int): Axis number
        """
        self.socket_put_confirmed(f"naxis={axis_Id_numeric}")
        self.socket_put_and_receive("XQ#NEWPAR")
        self.socket_put_confirmed("XQ#FRM")
        time.sleep(0.1)
        while self.is_axis_moving(None, axis_Id_numeric):
            time.sleep(0.1)

        if not self.axis_is_referenced(axis_Id_numeric):
            raise GalilError(f"Failed to find reference of axis {axis_Id_numeric}.")

        logger.info(f"Successfully found reference of axis {axis_Id_numeric}.")

    def show_running_threads(self) -> None:
        t = PrettyTable()
        t.title = f"Threads on {self.sock.host}:{self.sock.port}"
        t.field_names = [str(ax) for ax in range(self._axes_per_controller)]
        t.add_row(
            [
                "active" if self.is_thread_active(t) else "inactive"
                for t in range(self._axes_per_controller)
            ]
        )
        print(t)

    def is_motor_on(self, axis_Id) -> bool:
        return not bool(float(self.socket_put_and_receive(f"MG _MO{axis_Id}").strip()))

    def get_motor_limit_switch(self, axis_Id) -> list:
        """
        Get the status of the motor limit switches.

        Args:
            axis_Id (str): Axis identifier (e.g. 'A', 'B', 'C', ...)

        Returns:
            list: List of two booleans indicating if the low and high limit switch is active, respectively.
        """
        ret = self.socket_put_and_receive(f"MG _LR{axis_Id}, _LF{axis_Id}")
        low, high = ret.strip().split(" ")
        return [not bool(float(low)), not bool(float(high))]

    def describe(self) -> None:
        t = PrettyTable()
        t.title = f"{self.__class__.__name__} on {self.sock.host}:{self.sock.port}"
        t.field_names = [
            "Axis",
            "Name",
            "Connected",
            "Referenced",
            "Motor On",
            "Limits",
            "Position",
        ]
        for ax in range(self._axes_per_controller):
            axis = self._axis[ax]
            if axis is not None:
                t.add_row(
                    [
                        f"{axis.axis_Id_numeric}/{axis.axis_Id}",
                        axis.name,
                        axis.connected,
                        self.axis_is_referenced(axis.axis_Id_numeric),
                        self.is_motor_on(axis.axis_Id),
                        self.get_motor_limit_switch(axis.axis_Id),
                        axis.readback.read().get(axis.name).get("value"),
                    ]
                )
            else:
                t.add_row([None for t in t.field_names])
        print(t)

        self.show_running_threads()

    def galil_show_all(self) -> None:
        for controller in self._controller_instances.values():
            if isinstance(controller, GalilController):
                controller.describe()

    @staticmethod
    def axis_Id_to_numeric(axis_Id: str) -> int:
        return ord(axis_Id.lower()) - 97

    @staticmethod
    def axis_Id_numeric_to_alpha(axis_Id_numeric: int) -> str:
        return (chr(axis_Id_numeric + 97)).capitalize()


class GalilSignalBase(SocketSignal):
    def __init__(self, signal_name, **kwargs):
        self.signal_name = signal_name
        super().__init__(**kwargs)
        self.controller = self.parent.controller
        self.sock = self.parent.controller.sock


class GalilSignalRO(GalilSignalBase):
    def __init__(self, signal_name, **kwargs):
        super().__init__(signal_name, **kwargs)
        self._metadata["write_access"] = False

    def _socket_set(self, val):
        raise ReadOnlyError("Read-only signals cannot be set")


class GalilReadbackSignal(GalilSignalRO):
    @retry_once
    @threadlocked
    def _socket_get(self) -> float:
        """Get command for the readback signal

        Returns:
            float: Readback value after adjusting for sign and motor resolution.
        """

        current_pos = float(self.controller.socket_put_and_receive(f"TD{self.parent.axis_Id}"))
        current_pos *= self.parent.sign
        step_mm = self.parent.motor_resolution.get()
        return current_pos / step_mm

    def read(self):
        self._metadata["timestamp"] = time.time()
        val = super().read()
        if self.parent.axis_Id_numeric == 2:
            try:
                rt = self.parent.device_manager.devices[self.parent.rt]
                if rt.enabled:
                    rt.obj.controller.set_rotation_angle(val[self.parent.name]["value"])
            except KeyError:
                logger.warning("Failed to set RT value during readback.")
        return val


class GalilSetpointSignal(GalilSignalBase):
    setpoint = 0

    def _socket_get(self) -> float:
        """Get command for receiving the setpoint / target value.
        The value is not pulled from the controller but instead just the last setpoint used.

        Returns:
            float: setpoint / target value
        """
        return self.setpoint

    @retry_once
    @threadlocked
    def _socket_set(self, val: float) -> None:
        """Set a new target value / setpoint value. Before submission, the target value is adjusted for the axis' sign.
        Furthermore, it is ensured that all axes are referenced before a new setpoint is submitted.

        Args:
            val (float): Target value / setpoint value

        Raises:
            GalilError: Raised if not all axes are referenced.

        """
        target_val = val * self.parent.sign
        self.setpoint = target_val
        axes_referenced = self.controller.all_axes_referenced()
        if axes_referenced:
            while self.controller.is_thread_active(0):
                time.sleep(0.1)

            if self.parent.axis_Id_numeric == 2:
                angle_status = self.parent.device_manager.devices[
                    self.parent.rt
                ].obj.controller.feedback_status_angle_lamni()

                if angle_status:
                    self.controller.socket_put_confirmed("angintf=1")

            self.controller.socket_put_confirmed(f"naxis={self.parent.axis_Id_numeric}")
            self.controller.socket_put_confirmed(f"ntarget={target_val:.3f}")
            self.controller.socket_put_confirmed("movereq=1")
            self.controller.socket_put_confirmed("XQ#NEWPAR")
            while self.controller.is_thread_active(0):
                time.sleep(0.005)
        else:
            raise GalilError("Not all axes are referenced.")


class GalilMotorResolution(GalilSignalRO):
    @retry_once
    @threadlocked
    def _socket_get(self):
        return float(
            self.controller.socket_put_and_receive(f"MG stppermm[{self.parent.axis_Id_numeric}]")
        )


class GalilMotorIsMoving(GalilSignalRO):
    @threadlocked
    def _socket_get(self):
        return self.controller.is_axis_moving(self.parent.axis_Id, self.parent.axis_Id_numeric)

    def get(self):
        val = super().get()
        if val is not None:
            self._run_subs(
                sub_type=self.SUB_VALUE,
                value=val,
                timestamp=time.time(),
            )
        return val


class GalilAxesReferenced(GalilSignalRO):
    @threadlocked
    def _socket_get(self):
        return self.controller.all_axes_referenced()


class GalilMotor(Device, PositionerBase):
    USER_ACCESS = ["controller"]
    readback = Cpt(
        GalilReadbackSignal,
        signal_name="readback",
        kind="hinted",
    )
    user_setpoint = Cpt(GalilSetpointSignal, signal_name="setpoint")
    motor_resolution = Cpt(GalilMotorResolution, signal_name="resolution", kind="config")
    motor_is_moving = Cpt(GalilMotorIsMoving, signal_name="motor_is_moving", kind="normal")
    all_axes_referenced = Cpt(GalilAxesReferenced, signal_name="all_axes_referenced", kind="config")
    high_limit_travel = Cpt(Signal, value=0, kind="omitted")
    low_limit_travel = Cpt(Signal, value=0, kind="omitted")

    SUB_READBACK = "readback"
    SUB_CONNECTION_CHANGE = "connection_change"
    _default_sub = SUB_READBACK

    def __init__(
        self,
        axis_Id,
        prefix="",
        *,
        name,
        kind=None,
        read_attrs=None,
        configuration_attrs=None,
        parent=None,
        host="mpc2680.psi.ch",
        port=8081,
        limits=None,
        sign=1,
        socket_cls=SocketIO,
        device_manager=None,
        **kwargs,
    ):
        self.controller = GalilController(socket_cls=socket_cls, socket_host=host, socket_port=port)
        self.axis_Id = axis_Id
        self.controller.set_axis(axis=self, axis_nr=self.axis_Id_numeric)
        self.sign = sign
        self.tolerance = kwargs.pop("tolerance", 0.5)
        self.device_mapping = kwargs.pop("device_mapping", {})
        self.device_manager = device_manager

        if len(self.device_mapping) > 0 and self.device_manager is None:
            raise BECConfigError(
                "device_mapping has been specified but the device_manager cannot be accessed."
            )
        self.rt = self.device_mapping.get("rt")

        super().__init__(
            prefix,
            name=name,
            kind=kind,
            read_attrs=read_attrs,
            configuration_attrs=configuration_attrs,
            parent=parent,
            **kwargs,
        )
        self.readback.name = self.name
        self.controller.subscribe(
            self._update_connection_state, event_type=self.SUB_CONNECTION_CHANGE
        )
        self._update_connection_state()
        # self.readback.subscribe(self._forward_readback, event_type=self.readback.SUB_VALUE)

        if limits is not None:
            assert len(limits) == 2
            self.low_limit_travel.put(limits[0])
            self.high_limit_travel.put(limits[1])

    @property
    def limits(self):
        return (self.low_limit_travel.get(), self.high_limit_travel.get())

    @property
    def low_limit(self):
        return self.limits[0]

    @property
    def high_limit(self):
        return self.limits[1]

    def check_value(self, pos):
        """Check that the position is within the soft limits"""
        low_limit, high_limit = self.limits

        if low_limit < high_limit and not (low_limit <= pos <= high_limit):
            raise LimitError(f"position={pos} not within limits {self.limits}")

    def _update_connection_state(self, **kwargs):
        for walk in self.walk_signals():
            walk.item._metadata["connected"] = self.controller.connected

    def _forward_readback(self, **kwargs):
        kwargs.pop("sub_type")
        self._run_subs(sub_type="readback", **kwargs)

    @raise_if_disconnected
    def move(self, position, wait=True, **kwargs):
        """Move to a specified position, optionally waiting for motion to
        complete.

        Parameters
        ----------
        position
            Position to move to
        moved_cb : callable
            Call this callback when movement has finished. This callback must
            accept one keyword argument: 'obj' which will be set to this
            positioner instance.
        timeout : float, optional
            Maximum time to wait for the motion. If None, the default timeout
            for this positioner is used.

        Returns
        -------
        status : MoveStatus

        Raises
        ------
        TimeoutError
            When motion takes longer than `timeout`
        ValueError
            On invalid positions
        RuntimeError
            If motion fails other than timing out
        """
        self._started_moving = False
        timeout = kwargs.pop("timeout", 100)
        status = super().move(position, timeout=timeout, **kwargs)
        self.user_setpoint.put(position, wait=False)

        def move_and_finish():
            while self.motor_is_moving.get():
                logger.info("motor is moving")
                val = self.readback.read()
                self._run_subs(
                    sub_type=self.SUB_READBACK,
                    value=val,
                    timestamp=time.time(),
                )
                time.sleep(0.1)
            val = self.readback.read()
            success = np.isclose(
                val[self.name]["value"],
                position,
                atol=self.tolerance,
            )

            if not success:
                print(" stop")
            self._done_moving(success=success)
            logger.info("Move finished")

        threading.Thread(target=move_and_finish, daemon=True).start()
        try:
            if wait:
                status_wait(status)
        except KeyboardInterrupt:
            self.stop()
            raise

        return status

    @property
    def axis_Id(self):
        return self._axis_Id_alpha

    @axis_Id.setter
    def axis_Id(self, val):
        if isinstance(val, str):
            if len(val) != 1:
                raise ValueError(f"Only single-character axis_Ids are supported.")
            self._axis_Id_alpha = val
            self._axis_Id_numeric = self.controller.axis_Id_to_numeric(val)
        else:
            raise TypeError(f"Expected value of type str but received {type(val)}")

    @property
    def axis_Id_numeric(self):
        return self._axis_Id_numeric

    @axis_Id_numeric.setter
    def axis_Id_numeric(self, val):
        if isinstance(val, int):
            if val > 26:
                raise ValueError(f"Numeric value exceeds supported range.")
            self._axis_Id_alpha = self.controller.axis_Id_numeric_to_alpha(val)
            self._axis_Id_numeric = val
        else:
            raise TypeError(f"Expected value of type int but received {type(val)}")

    @property
    def egu(self):
        """The engineering units (EGU) for positions"""
        return "mm"

    def stage(self) -> List[object]:
        return super().stage()

    def unstage(self) -> List[object]:
        return super().unstage()

    def stop(self, *, success=False):
        self.controller.stop_all_axes()
        return super().stop(success=success)


if __name__ == "__main__":
    mock = False
    if not mock:
        leyey = GalilMotor("H", name="leyey", host="mpc2680.psi.ch", port=8081, sign=-1)
        leyey.stage()
        status = leyey.move(0, wait=True)
        status = leyey.move(10, wait=True)
        leyey.read()

        leyey.get()
        leyey.describe()

        leyey.unstage()
    else:
        from ophyd_devices.utils.socket import SocketMock

        leyex = GalilMotor(
            "G", name="leyex", host="mpc2680.psi.ch", port=8081, socket_cls=SocketMock
        )
        leyey = GalilMotor(
            "H", name="leyey", host="mpc2680.psi.ch", port=8081, socket_cls=SocketMock
        )
        leyex.stage()
        # leyey.stage()

        leyex.controller.galil_show_all()



import numpy as np
from ophyd import Component, Device, EpicsSignal, EpicsSignalRO


class XbpmCsaxsOp(Device):
    """Python wrapper for custom XBPMs in the cSAXS optics hutch

    This is  completely custom XBPM with templates directly in the
    VME repo. Thus it needs a custom ophyd template as well...

    WARN: The x and y are not updated by the IOC
    """

    sum = Component(EpicsSignalRO, "SUM", auto_monitor=True)
    x = Component(EpicsSignalRO, "POSH", auto_monitor=True)
    y = Component(EpicsSignalRO, "POSV", auto_monitor=True)
    s1 = Component(EpicsSignalRO, "CHAN1", auto_monitor=True)
    s2 = Component(EpicsSignalRO, "CHAN2", auto_monitor=True)
    s3 = Component(EpicsSignalRO, "CHAN3", auto_monitor=True)
    s4 = Component(EpicsSignalRO, "CHAN4", auto_monitor=True)


class XbpmBase(Device):
    """Python wrapper for X-ray Beam Position Monitors

    XBPM's consist of a metal-coated diamond window that ejects
    photoelectrons from the incoming X-ray beam. These electons
    are collected and their current is measured. Effectively
    they act as four quadrant photodiodes and are used as BPMs
    at the undulator beamlines of SLS.

    Note: EPICS provided signals are read only, but the user can
    change the beam position offset.
    """

    # Motor interface
    s1 = Component(EpicsSignalRO, "Current1", auto_monitor=True)
    s2 = Component(EpicsSignalRO, "Current2", auto_monitor=True)
    s3 = Component(EpicsSignalRO, "Current3", auto_monitor=True)
    s4 = Component(EpicsSignalRO, "Current4", auto_monitor=True)
    sum = Component(EpicsSignalRO, "SumAll", auto_monitor=True)
    asymH = Component(EpicsSignalRO, "asymH", auto_monitor=True)
    asymV = Component(EpicsSignalRO, "asymV", auto_monitor=True)
    x = Component(EpicsSignalRO, "X", auto_monitor=True)
    y = Component(EpicsSignalRO, "Y", auto_monitor=True)
    scaleH = Component(EpicsSignal, "PositionScaleX", auto_monitor=False)
    offsetH = Component(EpicsSignal, "PositionOffsetX", auto_monitor=False)
    scaleV = Component(EpicsSignal, "PositionScaleY", auto_monitor=False)
    offsetV = Component(EpicsSignal, "PositionOffsetY", auto_monitor=False)


class XbpmSim(XbpmBase):
    """Python wrapper for simulated X-ray Beam Position Monitors

    XBPM's consist of a metal-coated diamond window that ejects
    photoelectrons from the incoming X-ray beam. These electons
    are collected and their current is measured. Effectively
    they act as four quadrant photodiodes and are used as BPMs
    at the undulator beamlines of SLS.

    Note: EPICS provided signals are read only, but the user can
       change the beam position offset.

    This simulation device extends the basic proxy with a script that
    fills signals with quasi-randomized values.
    """

    # Motor interface
    s1w = Component(EpicsSignal, "Current1:RAW.VAL", auto_monitor=False)
    s2w = Component(EpicsSignal, "Current2:RAW.VAL", auto_monitor=False)
    s3w = Component(EpicsSignal, "Current3:RAW.VAL", auto_monitor=False)
    s4w = Component(EpicsSignal, "Current4:RAW.VAL", auto_monitor=False)
    rangew = Component(EpicsSignal, "RANGEraw", auto_monitor=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._MX = 0
        self._MY = 0
        self._I0 = 255.0
        self._x = np.linspace(-5, 5, 64)
        self._y = np.linspace(-5, 5, 64)
        self._x, self._y = np.meshgrid(self._x, self._y)

    def _simFrame(self):
        """Generator to simulate a jumping gaussian"""

        # define normalized 2D gaussian
        def gaus2d(x=0, y=0, mx=0, my=0, sx=1, sy=1):
            return np.exp(
                -((x - mx) ** 2.0 / (2.0 * sx**2.0) + (y - my) ** 2.0 / (2.0 * sy**2.0))
            )

        # Generator for dynamic values
        self._MX = 0.75 * self._MX + 0.25 * (10.0 * np.random.random() - 5.0)
        self._MY = 0.75 * self._MY + 0.25 * (10.0 * np.random.random() - 5.0)
        self._I0 = 0.75 * self._I0 + 0.25 * (255.0 * np.random.random())

        arr = self._I0 * gaus2d(self._x, self._y, self._MX, self._MY)
        return arr

    def sim(self):
        # Get next frame
        beam = self._simFrame()
        total = np.sum(beam)
        rnge = np.floor(np.log10(total) - 0.0)
        s1 = np.sum(beam[32:64, 32:64]) / 10**rnge
        s2 = np.sum(beam[0:32, 32:64]) / 10**rnge
        s3 = np.sum(beam[32:64, 0:32]) / 10**rnge
        s4 = np.sum(beam[0:32, 0:32]) / 10**rnge

        self.s1w.set(s1).wait()
        self.s2w.set(s2).wait()
        self.s3w.set(s3).wait()
        self.s4w.set(s4).wait()
        self.rangew.set(rnge).wait()
        # Print debug info
        print(f"Raw signals: R={rnge}\t{s1}\t{s2}\t{s3}\t{s4}")
        # plt.imshow(beam)
        # plt.show(block=False)
        # plt.pause(0.5)


# Automatically start simulation if directly invoked
if __name__ == "__main__":
    xbpm1 = XbpmSim("X01DA-FE-XBPM1:", name="xbpm1")
    xbpm2 = XbpmSim("X01DA-FE-XBPM2:", name="xbpm2")

    xbpm1.wait_for_connection(timeout=5)
    xbpm2.wait_for_connection(timeout=5)

    xbpm1.rangew.set(1).wait()
    xbpm2.rangew.set(1).wait()

    while True:
        print("---")
        xbpm1.sim()
        xbpm2.sim()
